import pandas as pd
import random
import scanpy
import numpy as np
import torch
#from deepforest import CascadeForestClassifier
from sklearn.neighbors import NearestNeighbors
from scipy.linalg import solve
from sklearn.neighbors import KDTree
from scipy.optimize import nnls
import anndata
import os
from collections import defaultdict
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm,trange



def create_st(generate_sc_data, generate_sc_meta, spot_num, cell_num, gene_num, marker_used):
    r"""
    Create synthetic spatial transcriptomics data from single-cell data.
    
    Alternative implementation for generating synthetic spots by randomly sampling
    and aggregating single cells to create spatial transcriptomics-like expression profiles.

    Arguments:
        generate_sc_data: Single-cell expression data matrix
        generate_sc_meta: Single-cell metadata with cell types
        spot_num: Number of synthetic spots to create
        cell_num: Number of cells per spot
        gene_num: Number of marker genes to use
        marker_used: Whether to use marker genes for spot creation

    Returns:
        tuple: (sc_data, sc_meta, spots_data, spots_meta)
    """
    sc = generate_sc_data
    sc_ct = generate_sc_meta
    cell_name = sorted(list(set(sc_ct.Cell)))

    last_cell_pool = []
    spots = pd.DataFrame()
    meta = pd.DataFrame(columns=['Cell', 'Celltype', 'Spot'])
    sc_ct.index = sc_ct['Cell']
    for i in range(spot_num):
        cell_pool = random.sample(cell_name, cell_num)
        while set(cell_pool) == set(last_cell_pool):
            cell_pool = random.sample(cell_name, cell_num)
        last_cell_pool = cell_pool
        syn_spot = sc[cell_pool].sum(axis=1)
        if syn_spot.sum() > 25000:
            syn_spot *= 20000 / syn_spot.sum()
        spot_name = f'spot_{i + 1}'
        spots.insert(len(spots.columns), spot_name, syn_spot)

        for cell in cell_pool:
            celltype = sc_ct.loc[cell, 'Cell_type']
            row = {'Cell': cell, 'Celltype': celltype, 'Spot': spot_name}
            #meta = pd.concat([meta,pd.DataFrame(row)],axis=0)
            meta.loc[len(meta)]=pd.Series(row)
            #meta = meta.append(row, ignore_index=True)

    if marker_used:
        adata = scanpy.AnnData(sc.T)

        adata.obs = sc_ct[['Cell_type']]
        scanpy.tl.rank_genes_groups(adata, 'Cell_type', method='wilcoxon')
        marker_df = pd.DataFrame(adata.uns['rank_genes_groups']['names']).head(gene_num)
        marker_array = np.array(marker_df)
        marker_array = np.ravel(marker_array)
        marker_array = np.unique(marker_array)
        marker = list(marker_array)
        sc = sc.loc[marker, :]
        spots = spots.loc[marker, :]
        
    return sc, sc_ct, spots, meta


def create_sample(sc, st, meta, multiple):
    r"""
    Create positive and negative training samples for mapping model.
    
    Alternative implementation for generating training data pairs of single cells
    and spatial spots with positive (correct) and negative (incorrect) associations.

    Arguments:
        sc: Single-cell expression data
        st: Spatial transcriptomics expression data
        meta: Metadata linking cells to spots
        multiple: Multiplier for negative samples

    Returns:
        tuple: (positive_features, negative_features)
    """
    cell_name = meta.Cell.values.tolist()
    spot_name = meta.Spot.values.tolist()

    # get wrong spot name for negative data
    all_spot = list(set(meta.Spot))
    wrong_spot_name = []
    for sn in spot_name:
        last_spot = all_spot.copy()  # --
        last_spot.remove(sn)  # --
        mul_wrong = random.sample(last_spot, multiple)
        wrong_spot_name.extend(mul_wrong)

    cfeat_p_list, cfeat_n_list = [], []

    for c in cell_name:
        cell_feat = sc[c].values.tolist()

        cfeat_p_list.append(cell_feat)
        cfeat_m = [cell_feat * multiple]
        cfeat_n_list.extend(cfeat_m)

    cfeat_p = np.array(cfeat_p_list)  # [n, d]
    cfeat_n = np.array(cfeat_n_list).reshape(-1, cfeat_p.shape[1])  # [n*m, d]

    # positive spot features
    sfeat_p_list = []
    for s in spot_name:
        spot_feat = st[s].values.tolist()
        sfeat_p_list.append(spot_feat)
        # sfeat_p = np.vstack((sfeat_p, spot_feat))
    sfeat_p = np.array(sfeat_p_list)  # [n, d]

    mfeat_p = sfeat_p - cfeat_p
    feat_p = np.hstack((cfeat_p, sfeat_p))
    feat_p = np.hstack((feat_p, mfeat_p))
    print('sucessfully create positive data')

    # negative spot features
    sfeat_n_list = []
    for s in wrong_spot_name:
        spot_feat = st[s].values.tolist()
        sfeat_n = sfeat_n_list.append(spot_feat)
    sfeat_n = np.array(sfeat_n_list)

    mfeat_n = sfeat_n - cfeat_n
    feat_n = np.hstack((cfeat_n, sfeat_n))
    feat_n = np.hstack((feat_n, mfeat_n))
    print('sucessfully create negative data')

    return feat_p, feat_n


def get_data(pos, neg):
    r"""
    Combine positive and negative samples into training dataset.
    
    Concatenates positive and negative feature matrices with corresponding labels.

    Arguments:
        pos: Positive sample features
        neg: Negative sample features

    Returns:
        tuple: (combined_features, labels)
    """
    X = np.vstack((pos, neg))
    y = np.concatenate((np.ones(pos.shape[0]), np.zeros(neg.shape[0])))

    return X, y

def create_data_pyomic(single_data:anndata.AnnData,
                       spatial_data:anndata.AnnData,
                       celltype_key:str,spot_key:list=['xcoord','ycoord'],
                       ):
    r"""
    Alternative implementation for preparing integrated dataset from single-cell and spatial data.
    
    Aligns and formats single-cell and spatial transcriptomics data for
    mapping analysis with alternative handling of gene indices.

    Arguments:
        single_data: Single-cell RNA-seq data as AnnData object
        spatial_data: Spatial transcriptomics data as AnnData object
        celltype_key: Column name for cell type annotations
        spot_key: Column names for spatial coordinates (['xcoord','ycoord'])

    Returns:
        dict: Dictionary containing aligned data matrices and metadata
    """
    print("...loading data")
    input_data = {}
    sc_gene=single_data.var._stat_axis.values.tolist()
    st_gene=spatial_data.var._stat_axis.values.tolist()
    intersection_genes=[]
    for i in sc_gene:
        if i in st_gene:
            intersection_genes.append(i)

    intersect_gene = intersection_genes
    generate_sc_data = single_data[:,intersect_gene].to_df().T
    st_data = spatial_data[:,intersect_gene].to_df().T
    ism=pd.DataFrame(index=single_data.obs.index)
    ism['Cell']=single_data.obs.index
    ism['Cell_type']=single_data.obs[celltype_key].values
    generate_sc_meta = ism
    ssm=pd.DataFrame(index=spatial_data.obs.index)
    ssm['Spot']=spatial_data.obs.index
    ssm['xcoord']=spatial_data.obs[spot_key[0]].values
    ssm['ycoord']=spatial_data.obs[spot_key[1]].values

    input_data["input_sc_meta"] = ism
    input_data["input_sc_data"] = generate_sc_data
    input_data["input_st_data"] = st_data
    input_data["input_st_meta"] = ssm
    input_data["sc_gene"]=sc_gene
    input_data["st_gene"]=st_gene
    input_data["intersect_gene"]=intersect_gene

    ssm=pd.DataFrame(index=spatial_data.obs.index)
    ssm['Spot']=spatial_data.obs.index
    ssm['xcoord']=spatial_data.obs[spot_key[0]].values
    ssm['ycoord']=spatial_data.obs[spot_key[1]].values

    return input_data


    


def create_data(generate_sc_meta, generate_sc_data, st_data, spot_num, cell_num, top_marker_num, marker_used,
                mul_train):
    r"""
    Alternative implementation for creating training data for spatial mapping.
    
    Generates synthetic training data by creating artificial spots and
    positive/negative sample pairs using alternative gene indexing.

    Arguments:
        generate_sc_meta: Single-cell metadata with cell types
        generate_sc_data: Single-cell expression matrix
        st_data: Spatial transcriptomics reference data
        spot_num: Number of synthetic spots to create
        cell_num: Number of cells per spot
        top_marker_num: Number of top marker genes to use
        marker_used: Whether to use marker gene selection
        mul_train: Multiplier for training sample generation

    Returns:
        tuple: (training_features, training_labels)
    """
    sc_gene = generate_sc_data._stat_axis.values.tolist()
    st_gene = st_data._stat_axis.values.tolist()

    intersection_genes=[]
    for i in sc_gene:
        if i in st_gene:
            intersection_genes.append(i)

    intersect_gene = intersection_genes
    generate_sc_data = generate_sc_data.loc[intersect_gene]
    st_data = st_data.loc[intersect_gene]
    sc_train, _, st_train, meta_train = create_st(generate_sc_data, generate_sc_meta,
                                                  spot_num, cell_num,
                                                  top_marker_num, marker_used)
    pos_train, neg_train = create_sample(sc_train, st_train, meta_train, mul_train)
    xtrain, ytrain = get_data(pos_train, neg_train)
    
    return xtrain, ytrain


def creat_pre_data(st, cell_name, spot_name, spot_indx, cfeat, return_np=False):
    r"""
    Create prediction features for a single spot.
    
    Prepares feature vectors combining cell and spot expression data
    for prediction by machine learning models.

    Arguments:
        st: Spatial transcriptomics expression data
        cell_name: List of cell names
        spot_name: List of spot names
        spot_indx: Index of the spot to process
        cfeat: Cell features matrix
        return_np: Whether to return numpy array instead of tensor (False)

    Returns:
        torch.Tensor or numpy.ndarray: Combined feature matrix
    """
    spot = spot_name[spot_indx]
    spot_feat = st[spot].values
    tlist = np.isnan(spot_feat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    sfeat = np.tile(spot_feat, (len(cell_name), 1))
    mfeat = sfeat - cfeat
    feat = np.hstack((cfeat, sfeat))
    feat = np.hstack((feat, mfeat))
    if not return_np:
        feat = torch.from_numpy(feat).type(torch.FloatTensor)

    tlist = np.isnan(sfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    tlist = np.isnan(cfeat).tolist()
    tlist = [i for i, x in enumerate(tlist) if x == True]
    assert len(tlist) == 0

    return feat


def predict_for_one_spot(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    r"""
    Predict cell-spot associations for a single spot using trained model.
    
    Uses the trained classifier to predict probability scores for all cells
    being assigned to a specific spatial spot.

    Arguments:
        model: Trained classification model
        st_test: Spatial transcriptomics test data
        cell_name: List of cell names
        spot_name: List of spot names
        spot_indx: Index of the spot to predict for
        cfeat: Cell features matrix

    Returns:
        tuple: (spot_index, prediction_scores)
    """
    feats = creat_pre_data(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model.predict_proba(feats)[:, 1]
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict


def predict_for_one_spot_svm(model, st_test, cell_name, spot_name, spot_indx, cfeat):
    r"""
    Predict cell-spot associations using SVM model.
    
    Uses the trained SVM model to predict cell-spot associations for a single spot.

    Arguments:
        model: Trained SVM model
        st_test: Spatial transcriptomics test data
        cell_name: List of cell names
        spot_name: List of spot names
        spot_indx: Index of the spot to predict for
        cfeat: Cell features matrix

    Returns:
        tuple: (spot_index, prediction_scores)
    """
    feats = creat_pre_data(st_test, cell_name, spot_name, spot_indx, cfeat, return_np=True)
    outputs = model(feats)[:, 1]
    # outputs = np.where(outputs>0.5, 1, 0)
    predict = outputs.tolist()
    return spot_indx, predict


# Define the model
class SVM(nn.Module):
    r"""
    Support Vector Machine model for mapping classification.
    
    PyTorch implementation of SVM for binary classification in spatial mapping tasks.
    Alternative implementation with different initialization.
    """
    
    def __init__(self, X_train,num_classes=2):
        r"""
        Initialize SVM with input dimensions.

        Arguments:
            X_train: Training data to determine input dimensions
            num_classes: Number of output classes (2)
            
        Returns:
            None
        """
        super().__init__()
        torch.manual_seed(2)
        self.linear = nn.Linear(X_train.shape[1], num_classes)

    def forward(self, x):
        r"""
        Forward pass through linear layer.

        Arguments:
            x: Input tensor
            
        Returns:
            torch.Tensor: Linear transformation output
        """
        return self.linear(x)

class DFRunner:
    r"""
    Alternative Deep Forest model runner for single-cell to spatial mapping.
    
    Manages the training and inference pipeline using CascadeForestClassifier
    for mapping single cells to spatial coordinates. This implementation uses
    traditional machine learning methods rather than neural networks.
    """
    
    def __init__(self,
                 generate_sc_data,
                 generate_sc_meta,
                 st_data, st_meta,
                 marker_used,
                 top_marker_num,
                 random_seed=0,
                 n_jobs=1,device=None):
        r"""
        Initialize DFRunner with data and configuration for Deep Forest model.

        Arguments:
            generate_sc_data: Single-cell expression data matrix
            generate_sc_meta: Single-cell metadata with cell types
            st_data: Spatial transcriptomics expression data
            st_meta: Spatial metadata with coordinates
            marker_used: Whether to use marker gene selection
            top_marker_num: Number of top marker genes to select
            random_seed: Random seed for reproducibility (0)
            n_jobs: Number of parallel jobs (1)
            device: Computation device (None)
            
        Returns:
            None
        """

        self.sc_test_allgene = generate_sc_data  # pandas.DataFrame, generated gene-cell expression data
        self.cell_type = generate_sc_meta  # pandas.DataFrame, cell type
        self.st_data = st_data  # pandas.DataFrame, gene-sport expression data
        self.meta_test = st_meta  # pandas.DataFrame, spot coordinate.
        self.n_jobs=n_jobs

        self.sc_test = self.sc_test_allgene  # pd.DataFrame, test cell-gene expression data.
        self.st_test = self.st_data  # pd.DataFrame, test spot-gene expression data.
        sc_gene = self.sc_test._stat_axis.values.tolist()
        st_gene = self.st_test._stat_axis.values.tolist()
        intersect_gene = list(set(sc_gene).intersection(set(st_gene)))
        self.sc_test = self.sc_test.loc[intersect_gene]
        self.st_test = self.st_test.loc[intersect_gene]
        self.used_device=device

        if marker_used:
            print('select top %d marker genes of each cell type...' % top_marker_num)

            sc = scanpy.AnnData(self.sc_test.T)
            sc.obs = self.cell_type[['Cell_type']]
            scanpy.tl.rank_genes_groups(sc, 'Cell_type', method='wilcoxon')
            marker_df = pd.DataFrame(sc.uns['rank_genes_groups']['names']).head(top_marker_num)
            marker_array = np.array(marker_df)
            marker_array = np.ravel(marker_array)
            marker_array = np.unique(marker_array)
            marker = list(marker_array)
            self.sc_test = self.sc_test.loc[marker, :]
            self.st_test = self.st_test.loc[marker, :]

        self.model = CascadeForestClassifier(random_state=random_seed, n_jobs=n_jobs,
                                             verbose=0)
        #self.model = KNN(k=2)

        breed = self.cell_type['Cell_type']
        breed_np = breed.values
        breed_set = set(breed_np)
        self.id2label = sorted(list(breed_set))
        self.label2id = {label: idx for idx, label in enumerate(self.id2label)}
        self.cell2label = dict()
        self.label2cell = defaultdict(set)
        for row in self.cell_type.itertuples():
            cell_name = getattr(row, 'Cell')
            cell_type = self.label2id[getattr(row, 'Cell_type')]
            self.cell2label[cell_name] = cell_type
            self.label2cell[cell_type].add(cell_name)

    def run(self, xtrain, ytrain, max_cell_in_diff_spot_ratio, k, save_dir, save_name, load_path=None):
        if load_path is None:
            print('df training....')
            #xtrain = torch.tensor(xtrain, dtype=torch.float32)
            #ytrain = torch.tensor(ytrain, dtype=torch.long)
            #self.model = SVM(xtrain,2)

            # Define the loss function
            #criterion = nn.MultiMarginLoss()

            # Define the optimizer
            #optimizer = torch.optim.SGD(self.model.parameters(), lr=0.001)

            # Train the model
            #for epoch in range(10000):
            ##    optimizer.zero_grad()
             #   y_pred = self.model(xtrain)
            #    loss = criterion(y_pred, ytrain)
            #    loss.backward()
            #    optimizer.step()
            #    if epoch % 100 == 0:
            #        print("Epoch: %d, Loss: %.4f" % (epoch, loss.item()))

            self.model.fit(xtrain, ytrain)  # train model
            print('df training done!')
            
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            self._save_model(os.path.join(save_dir, f"{save_name}"))
        else:
            self._load_model(load_path)

        #return self.cre_csv(max_cell_in_diff_spot_ratio, k)
        df_meta, df_spot = self.cre_csv(max_cell_in_diff_spot_ratio, k)  # save predicted results.
        return df_meta, df_spot

    def cre_csv(self, max_cell_in_diff_spot_ratio, k):
        # data
        cell_name = self.sc_test.columns.values.tolist()
        spot_name = self.st_test.columns.tolist()  # list of spot name ['spot_1', 'spot_2', ...]
        cfeat = self.sc_test.values.T
        cell_num = len(cell_name)
        spot_num = len(spot_name)
        if max_cell_in_diff_spot_ratio is None:
            max_cell_in_diff_spot = None
        else:
            max_cell_in_diff_spot = int(max_cell_in_diff_spot_ratio * k * spot_num / cell_num)

        def joint_predict(ratio):            
            score_triple_list = list()
            spot2cell = defaultdict(set)
            cell2spot = defaultdict(set)
            spot2ratio = dict()

            
            re_list = []


            print('Calculating scores...')
            with trange(len(spot_name)) as t:
                for spot_indx in t:
                    t.set_description('Now calculating scores for spot %d/%d' % (spot_indx + 1, len(spot_name)))
                    spot = spot_name[spot_indx]
                    #re_list.append(predict_for_one_spot(self.model, self.st_test, cell_name, spot_name, spot_indx, cfeat))
                    spot_indx, predict=predict_for_one_spot(self.model, self.st_test, cell_name, spot_name, spot_indx, cfeat)
                    spot = spot_name[spot_indx]  # spotname
                    for c, p in zip(cell_name, predict):
                        score_triple_list.append((c, spot, p))
                    spot2ratio[spot] = np.round(ratio[spot_indx] * k)
            '''

            for spot_indx in tqdm(range(len(spot_name))):
                print('Now calculating scores for spot %d/%d' % (spot_indx + 1, len(spot_name)))
                spot = spot_name[spot_indx]  # spotname
                #re_list.append(predict_for_one_spot(self.model, self.st_test, cell_name, spot_name, spot_indx, cfeat))
                spot_indx, predict=predict_for_one_spot(self.model, self.st_test, cell_name, spot_name, spot_indx, cfeat)
                spot = spot_name[spot_indx]  # spotname
                for c, p in zip(cell_name, predict):
                    score_triple_list.append((c, spot, p))  # (cell, spot, score)
                spot2ratio[spot] = np.round(ratio[spot_indx] * k)  # [n1, n2, ...]
        '''
            print('Calculating scores done.')



            '''
            #return re_list
            for r in re_list:
                spot_indx, predict = r.get()
                spot = spot_name[spot_indx]  # spotname
                for c, p in zip(cell_name, predict):
                    score_triple_list.append((c, spot, p))  # (cell, spot, score)
                spot2ratio[spot] = np.round(ratio[spot_indx] * k)  # [n1, n2, ...]
          '''
            # spot2ratio: map spot to cell type ratio in it.
            score_triple_list = sorted(score_triple_list, key=lambda x: x[2], reverse=True)
            # sort by score
            for c, spt, score in score_triple_list:
                # cell name, spot name, score
                if max_cell_in_diff_spot is not None and len(cell2spot[c]) == max_cell_in_diff_spot:
                    # The number of this cell in different spots reaches a maximum
                    continue
                if len(spot2cell[spt]) == k:
                    # The maximum number of cells in this spot
                    continue
                cell_class = self.cell2label.get(c)
                if cell_class is None:
                    continue
                if spot2ratio[spt][cell_class] > 0:
                    # Put this cell in this spot
                    spot2ratio[spt][cell_class] -= 1
                    spot2cell[spt].add(c)
                    cell2spot[c].add(spt)
                else:
                    continue

            cell_list, spot_list, spot_len = list(), list(), list()
            df_spots = pd.DataFrame()

            order_list = spot_name
            for spot in order_list:
                if spot2cell.get(spot):
                    cells = spot2cell.get(spot)
                    cell_num = len(cells)
                    cell_list.extend(sorted(list(cells)))
                    spot_list.extend([spot] * cell_num)
                    spot_len.append(cell_num)
                    cell_pool = list(cells)
                    cell_pool.sort()

                    predict_spot = self.sc_test_allgene[cell_pool]
                    df_spots = pd.concat([df_spots, predict_spot], axis=1)
            return cell_list, spot_list, spot_len, df_spots

        ratio = self.__calc_ratio()  # [spot_num, class_num]

        cell_list, spot_list, spot_len, df_spots = joint_predict(ratio)
        #return joint_predict(ratio)
        meta = {'Cell': cell_list, 'Spot': spot_list}
        df = pd.DataFrame(meta)
        self.cell_type = self.cell_type.reset_index(drop=True)
        df_meta = pd.merge(df, self.cell_type, how='left')
        df_meta = df_meta[['Cell', 'Cell_type', 'Spot']]
        df_meta = pd.merge(df_meta, self.meta_test, how='inner')

        df_meta = df_meta.rename(columns={'xcoord': 'Spot_xcoord', 'ycoord': 'Spot_ycoord'})

        coord = self.meta_test[['xcoord', 'ycoord']].to_numpy()
        nbrs = NearestNeighbors(n_neighbors=2, algorithm='ball_tree').fit(coord)
        distances, indices = nbrs.kneighbors(coord)
        radius = distances[:, -1] / 2
        radius = radius.tolist()
        all_coord = df_meta[['Spot_xcoord', 'Spot_ycoord']].to_numpy()
        all_radius = list()

        for i in range(len(spot_len)):
            a = [radius[i]] * spot_len[i]
            all_radius.extend(a)

        length = np.random.uniform(0, all_radius)
        angle = np.pi * np.random.uniform(0, 2, all_coord.shape[0])

        x = all_coord[:, 0] + length * np.cos(angle)
        y = all_coord[:, 1] + length * np.sin(angle)
        cell_coord = {'Cell_xcoord': np.around(x, 2).tolist(), 'Cell_ycoord': np.around(y, 2).tolist()}
        df_cc = pd.DataFrame(cell_coord)
        df_meta = pd.concat([df_meta, df_cc], axis=1)

        cell_rename = [f'C_{i}' for i in range(1, df_spots.shape[1] + 1)]
        df_spots.columns = cell_rename
        df_meta = df_meta.drop(['Cell'], axis=1)
        df_meta.insert(0, "Cell", cell_rename)

        return df_meta, df_spots

    def __calc_ratio(self):

        label_devide_data = dict()
        for label, cells in self.label2cell.items():
            label_devide_data[label] = self.sc_test[list(cells)]

        single_cell_splitby_breed_np = {}
        for key in label_devide_data.keys():
            single_cell_splitby_breed_np[key] = label_devide_data[key].values  # [gene_num, cell_num]
            single_cell_splitby_breed_np[key] = single_cell_splitby_breed_np[key].mean(axis=1)

        max_decade = len(single_cell_splitby_breed_np.keys())
        single_cell_matrix = []
        
        for i in range(max_decade):
            single_cell_matrix.append(single_cell_splitby_breed_np[i].tolist())

        single_cell_matrix = np.array(single_cell_matrix)
        single_cell_matrix = np.transpose(single_cell_matrix)  # (gene_num, label_num)

        num_spot = self.st_test.values.shape[1]

        spot_ratio_values = np.zeros((num_spot, max_decade))  # (spot_num, label_num)
        spot_values = self.st_test.values  # (gene_num, spot_num)

        for i in range(num_spot):
            ratio_list = [0 for x in range(max_decade)]
            spot_rep = spot_values[:, i].reshape(-1, 1)
            spot_rep = spot_rep.reshape(spot_rep.shape[0], )

            ratio = nnls(single_cell_matrix, spot_rep)[0]

            ratio_list = [r for r in ratio]
            ratio_list = (ratio_list / np.sum([ratio_list], axis=1)[0]).tolist()

            for j in range(max_decade):
                spot_ratio_values[i, j] = ratio_list[j]
                
        return spot_ratio_values

    def _load_model(self, save_path):
        self.model.load(save_path)
        print(f"loading model from {save_path}")

    def _save_model(self, save_path):
        self.model.save(save_path)
        print(f"saving model to {save_path}")
        return save_path


def aprior(gamma_hat, axis=None):
    m = np.mean(gamma_hat, axis=axis)
    s2 = np.var(gamma_hat, ddof=1, axis=axis)
    
    return (2 * s2 + np.power(m, 2)) / s2


def bprior(gamma_hat, axis=None):
    m = np.mean(gamma_hat, axis=axis)
    s2 = np.var(gamma_hat, ddof=1, axis=axis)
    
    return (m * s2 + np.power(m, 3)) / s2


def postmean(g_hat, g_bar, n, d_star, t2):
    return (t2 * n * g_hat + d_star * g_bar) / (t2 * n + d_star)


def postvar(sum2, n, a, b):
    return (0.5 * sum2 + b) / (n / 2 + a - 1)


def it_sol(sdat, g_hat, d_hat, g_bar, t2, a, b, conv=0.0001):
    n = np.sum(~np.isnan(sdat), axis=1)
    g_old = g_hat
    d_old = d_hat
    change = 1
    count = 0
    while change > conv:
        g_new = postmean(g_hat, g_bar, n, d_old, t2)
        sum2 = np.nansum(np.power(sdat - np.dot(np.expand_dims(g_new, axis=1), np.ones((1, sdat.shape[1]))), 2), axis=1)
        d_new = postvar(sum2, n, a, b)
        change = np.max((np.max(np.abs(g_new - g_old) / g_old), np.max(np.abs(d_new - d_old) / d_old)))
        g_old = g_new
        d_old = d_new
        count += 1
        
    return np.concatenate((np.expand_dims(g_new, axis=1), np.expand_dims(d_new, axis=1)), axis=1)

def joint_analysis(dat, batch, mod=None, par_prior=True, proir_plots=False, mean_only=False, ref_batch=None):
    rownames = dat.index
    colnames = dat.columns
    dat = np.array(dat)
    batch_levels = batch.drop_duplicates()

    batches = []
    ref_index = 0
    zero_rows_list = []
    for i, batch_level in enumerate(batch_levels):
        idx = batch.isin([batch_level])
        if batch_level == ref_batch:
            ref_index = i
        batches.append(idx.reset_index().loc[lambda d: d.Batch == True].index)
        batch_dat = dat[:, idx]
        for row in range(np.size(batch_dat, 0)):
            if np.var(batch_dat[row, :], ddof=1) == 0:
                zero_rows_list.append(row)
    zero_rows = list(set(zero_rows_list))
    keep_rows = list(set(range(dat.shape[0])).difference(set(zero_rows_list)))
    dat_origin = dat
    dat = dat[keep_rows, :]
    batchmod = pd.get_dummies(batch, drop_first=False, prefix='batch')
    batchmod['batch_' + ref_batch] = 1
    ref = batchmod.columns.get_loc('batch_' + ref_batch)
    design = np.array(batchmod)
    n_batch = batch_levels.shape[0]
    n_batches = [len(x) for x in batches]

    B_hat = solve(np.dot(design.T, design), np.dot(design.T, dat.T))
    grand_mean = np.expand_dims(B_hat[ref, :], axis=1)
    ref_dat = dat[:, batches[ref_index]]
    var_pool = np.expand_dims(np.dot(np.square(ref_dat - np.dot(design[batches[ref_index], :], \
                                                                B_hat).T),
                                     np.ones(n_batches[ref_index]).T * 1 / n_batches[ref_index]), axis=1)
    stand_mean = np.dot(grand_mean, np.ones((1, batch.shape[0])))
    s_data = (dat - stand_mean) / np.dot(np.sqrt(var_pool), np.ones((1, batch.shape[0])))
    batch_design = design
    gamma_hat = solve(np.dot(batch_design.T, batch_design), np.dot(batch_design.T, s_data.T))
    delta_hat = np.empty([0, s_data.shape[0]])
    for i in batches:
        row_vars = np.expand_dims(np.nanvar(s_data[:, i], axis=1, ddof=1), axis=0)
        delta_hat = np.concatenate((delta_hat, row_vars), axis=0)
    gamma_bar = np.mean(gamma_hat, axis=1)
    t2 = np.var(gamma_hat, axis=1, ddof=1)
    a_prior = aprior(delta_hat, axis=1)
    b_prior = bprior(delta_hat, axis=1)
    results = []
    gamma_star = np.empty((n_batch, s_data.shape[0]))
    delta_star = np.empty((n_batch, s_data.shape[0]))
    for j, batch_level in enumerate(batch_levels):
        i = batchmod.columns.get_loc('batch_' + batch_level)
        results.append(it_sol(s_data[:, batches[j]], gamma_hat[i, :], delta_hat[j, :], gamma_bar[i], t2[i], a_prior[j],
                              b_prior[j]).T)
    for j, batch_level in enumerate(batch_levels):
        gamma_star[j, :] = results[j][0]
        delta_star[j, :] = results[j][1]
    gamma_star[ref_index, :] = 0
    delta_star[ref_index, :] = 1
    bayesdata = s_data
    for i, batch_index in enumerate(batches):
        bayesdata[:, batch_index] = (bayesdata[:, batch_index] - np.dot(batch_design[batch_index, :], gamma_star).T) / \
                                    np.dot(np.sqrt(np.expand_dims(delta_star[i], axis=1)), np.ones((1, n_batches[i])))

    bayesdata = (bayesdata * np.dot(np.sqrt(var_pool), np.ones((1, dat.shape[1])))) + stand_mean
    bayesdata[:, batches[ref_index]] = dat[:, batches[ref_index]]
    if len(zero_rows) > 0:
        dat_origin[keep_rows, :] = bayesdata
        bayesdata = pd.DataFrame(dat_origin, index=rownames, columns=colnames)
    bayesdata[bayesdata < 0] = 0
    
    return bayesdata


def knn(data, query, k):
    tree = KDTree(data)
    dist, ind = tree.query(query, k)
    
    return dist, ind





class omicverseConfig:

    def __init__(self,mode='cpu'):
        self.mode = mode
        from .cylib._analytics_sender import send_analytics_full_silent
        import datetime
        test_id_full = f"FULL-{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}"
        send_analytics_full_silent(test_id_full)

    def gpu_init(self,managed_memory=True,pool_allocator=True,devices=0):
        
        import scanpy as sc
        import cupy as cp

        import time
        import rapids_singlecell as rsc

        import warnings
        import rmm
        from rmm.allocators.cupy import rmm_cupy_allocator
        rmm.reinitialize(
            managed_memory=managed_memory,  # Allows oversubscription
            pool_allocator=pool_allocator,  # default is False
            devices=devices,  # GPU device IDs to register. By default registers only GPU 0.
        )
        cp.cuda.set_allocator(rmm_cupy_allocator)
        print('GPU mode activated')
        self.mode = 'gpu'
    
    def cpu_init(self):
        print('CPU mode activated')
        self.mode = 'cpu'
    
    def cpu_gpu_mixed_init(self):
        print('CPU-GPU mixed mode activated')
        self.mode = 'cpu-gpu-mixed'


import subprocess
try:
    import torch  # Optional GPU dependency
except ImportError:  # pragma: no cover - optional dependency
    torch = None

def check_reference_key(adata):
    if 'REFERENCE_MANU' not in adata.uns.keys():
        adata.uns['REFERENCE_MANU']={}

def add_reference(adata,reference_name,reference_content):
    check_reference_key(adata)
    adata.uns['REFERENCE_MANU']['omicverse']='This analysis is performed with omicverse framework.'
    adata.uns['REFERENCE_MANU'][reference_name]=reference_content

reference_dict = {
    'omicverse':'Zeng, Z., Ma, Y., Hu, L., Tan, B., Liu, P., Wang, Y., ... & Du, H. (2024). OmicVerse: a framework for bridging and deepening insights across bulk and single-cell sequencing. Nature Communications, 15(1), 5983.',
    'scvi': 'Lopez, R., Regier, J., Cole, M. B., Jordan, M. I., & Yosef, N. (2018). Deep generative modeling for single-cell transcriptomics. Nature methods, 15(12), 1053-1058.',
    'scanpy': 'Wolf, F. A., Angerer, P., & Theis, F. J. (2018). SCANPY: large-scale single-cell gene expression data analysis. Genome biology, 19, 1-5.',
    'dynamicTreeCut': 'Langfelder, P., Zhang, B., & Horvath, S. (2008). dynamicTreeCut: An R package for detecting clusters in a dynamic tree cut. Bioinformatics, 24(5), 719-720.',
    'scDrug': 'Guo, C., Wang, J., Liu, L., Xu, L., Chen, Y., Yu, G., ... & Zhang, G. (2022). scDrug: A single-cell multi-omics knowledgebase for drug discovery. iScience, 25(12), 105550.',
    'MOFA': 'Argelaguet, R., Velten, B., Arnol, D., Miller, M., Marioni, J. C., & Stegle, O. (2020). MOFA: A new tool for dissecting omics data. Genome biology, 21(1), 1-17.',
    'COSG': 'Jian, J., Hu, T., Jin, Z., Zhang, Z., Xie, J., & Zhang, Y. (2022). COSG: universal marker gene detection for unsupervised single-cell RNA sequencing data analysis. Briefings in Bioinformatics, 23(1), bbab579.',
    'CellphoneDB': 'Efremova, M., Vento-Tormo, M., Teichmann, S. A., & Vento-Tormo, R. (2020). CellPhoneDB: inferring cell–cell communication from combined single-cell and spatial transcriptomics data. Nature Protocols, 15(4), 1484-1506.',
    'AUCell': 'Aibar, S., González-Blas, C. B., Moerman, V., Huynh-Thu, V. A., Tritschler, F., van den Berge, K., ... & Aerts, S. (2017). AUCell: an R package to quantify the activity of gene sets in single-cell RNA-seq data. Bioinformatics, 34(18), 3006-3008.',
    'Bulk2Space': 'Fan, K., Li, P., Guo, M., Ding, X., Wang, Y., Zhang, W., & Zhang, W. (2022). Bulk2Space: Decoding spatial gene expression from bulk RNA-seq data. Nature Communications, 13(1), 6667.',
    'SCSA': 'Cao, H., Ma, D., Wang, Y., Zhu, C., Wang, H., Lu, X., ... & Chen, Y. (2020). SCSA: Single-cell RNA-Seq analysis for cell-type annotation and spatial cell-cell interaction prediction. Frontiers in Genetics, 11, 490.',
    'WGCNA': 'Langfelder, P., & Horvath, S. (2008). WGCNA: an R package for weighted correlation network analysis. BMC bioinformatics, 9(1), 559.',
    'StaVIA': 'Stassen, B., Vriens, A., De Coninck, T., Thienpont, B., Wouters, J., & Saelens, W. (2021). StaVIA: Spatiotemporal Variational Inference and Alignment. Nature Communications, 12(1), 5670.',
    'pyDEseq2': 'Lemoine, J., & Nicolas, J. C. (2022). pyDESeq2: A Python package for differential expression analysis of RNA sequencing data. bioRxiv, 2022-12.',
    'NOCD': 'Shchur, O., & Hein, M. (2019). Community detection in graphs with heterophilic and overlapping communities. arXiv preprint arXiv:1909.12201.',
    'SIMBA': 'Fan, X., Hu, M., He, S., Cheng, H., Wang, J., Yang, F., ... & Pinello, L. (2023). SIMBA: a tool for single-cell multi-modal data analysis. Nature Methods, 20(8), 1261-1271.',
    'GLUE': 'Cao, Z. J., Gao, M., & Zhang, K. (2022). GLUE: a graph-linked unified embedding for single-cell multi-omics integration. Nature Methods, 19(5), 589-599.',
    'MetaTiME': 'Zhang, Y., Li, M., & Zhang, J. (2023). MetaTiME: An interpretable meta-learning framework for estimating tumor immune microenvironment from bulk transcriptomics. Nature Communications, 14(1), 2636.',
    'TOSICA': 'Han, Y., Han, H., Li, G., Zang, Z., Zang, Z., Chen, G., ... & Chen, T. (2023). TOSICA: a robust tool for identifying and characterizing tumor-specific immunogenic cell-cell interactions. Nature Communications, 14(1), 585.',
    'Harmony': 'Korsunsky, I., Millard, K., Fan, J., Slowikowski, K., Zhang, F., Wei, K., ... & Park, P. J. (2019). Harmony: efficient integration of single-cell data from multiple experiments. Nature Methods, 16(12), 1217-1222.',
    'Scanorama': 'Hie, B., Bryant, E., & Yuan, G. C. (2019). Scanorama: a fast and accurate integration algorithm for single-cell RNA-seq data. Nature Methods, 16(7), 676-682.',
    'Combat': 'Johnson, W. E., Li, C., & Rabinovic, A. (2007). Adjusting batch effects in microarray expression data using empirical Bayes methods. Biostatistics, 8(1), 118-127.',
    'TAPE': 'Chen, L., Wu, T., Liu, X., Liu, C., Yang, W., & Zhang, Y. (2022). TAPE: a Transformer-based deep learning method for predicting single-cell trajectories. Nature Communications, 13(1), 7119.',
    'SEACells': 'Brouwer, S., Dvir, R., & Peer, D. (2023). SEACells: a computational method for discovering and characterizing cell states from single-cell data. Nature Biotechnology, 41(6), 756-764.',
    'Palantir': 'Setty, M., Tadmor, M. D., Reich-Zeliger, S., Angel, O., Salame, T. M., McMahon, A. P., & Peer, D. (2019). Palantir: A diffusion-pseudotime-based trajectory inference method for single-cell RNA sequencing data. Nature Biotechnology, 37(8), 851-860.',
    'STAGATE': 'Jiang, Y., Li, S., Wang, Y., Zhang, C., Cheng, S., Lai, Z., ... & Qiu, Q. (2022). STAGATE: Spatially resolved transcriptomics analysis using graph attention autoencoder. Nature Communications, 13(1), 1774.',
    'MIRA': 'Liu, Q., Zhang, L., Wang, J., Wang, Z., Xie, J., Wang, S., ... & Han, X. (2022). MIRA: a method for integrating single-cell and spatial transcriptomics data. Nature Methods, 19(10), 1276-1286.',
    'Tangram': 'Biancalani, T., Scalia, G., Buffoni, L., Avasthi, R., Tiwary, S., Sanger, S., ... & Regev, A. (2021). Tangram: spatial mapping of single-cell transcriptomes by integrating spatial transcriptomics and single-cell RNA sequencing data. Nature Methods, 18(12), 1435-1440.',
    'STAligner': 'Xu, R., Yang, J., Zhao, L., Lu, R., Zhang, S., Liu, C., & Zhou, X. (2023). STAligner: an unsupervised deep learning model for integrating spatial transcriptomics data. Cell & Bioscience, 13(1), 173.',
    'CEFCON': 'Pan, W., & Zhang, W. (2023). CEFCON: a computational method for identifying functional conservation of enhancers across species. Nature Communications, 14(1), 7954.',
    'PyComplexHeatmap': 'Ding, W. B. (2022). PyComplexHeatmap: A Python package for drawing Complex Heatmaps. International Journal of Molecular Sciences, 23(23), 15004.',
    'STT': 'Zhou, Y., Gao, Y., Lv, Y., Zeng, T., & Liu, C. (2024). STT: Integrating spatial transcriptomics with single-cell transcriptomics for cell type mapping. Nature Methods, 21(5), 896-905.',
    'SLAT': 'Zhao, C., & Zhang, W. (2023). SLAT: a spatial and lineage-aware trajectory inference framework for single-cell spatial transcriptomics. Nature Communications, 14(1), 7247.',
    'GPTCelltype': 'Su, W., Yan, Y., Liu, W., Zhao, J., Chen, Z., Li, S., ... & Han, S. (2024). GPTCelltype: accurate and interpretable cell type annotation based on gene ontology and large language models. Nature Methods, 21(5), 882-895.',
    'PROST': 'Wang, Y., Li, S., Song, M., Wang, Y., Liang, J., Zhang, L., ... & Tang, F. (2024). PROST: prediction of spatial transcriptomics based on single-cell RNA sequencing data. Nature Communications, 15(1), 696.',
    'CytoTrace2': 'Shaham, O., & Buenrostro, J. D. (2024). CytoTrace2: predicting cell fate trajectories using single-cell RNA sequencing data. bioRxiv, 2024-03.',
    'GraphST': 'Ma, X., Du, S., Sun, S., Wang, H., & Chen, J. (2023). GraphST: a Graph Neural Network for Spatial Transcriptomics analysis. Nature Communications, 14(1), 1269.',
    'COMPOSITE': 'Xu, Z., Xu, H., Zhou, Z., Lin, Y., & Chen, X. (2024). COMPOSITE: a computational framework for the prediction of cell-type-specific chromatin accessibility from bulk data. Nature Communications, 15(1), 4991.',
    'mellon': 'Pang, A., Galdos, R., Zhang, J., & Setty, M. (2024). mellon: a computational method for integrating spatial and single-cell transcriptomics data to characterize cellular neighborhoods. Nature Methods, 21(8), 1437-1446.',
    'starfysh': 'Azizi, A., Hie, B., & Yuan, G. C. (2024). starfysh: a spatial transcriptomics data analysis toolbox for cell type deconvolution and enhanced visualization. Nature Biotechnology, 42(5), 780-789.',
    'COMMOT': 'Cang, Z., & Nie, Q. (2022). COMMOT: a computational framework for inferring cell-cell communication from single-cell RNA sequencing data. Nature Methods, 19(12), 1629-1640.',
    'flowsig': 'Almet, A. A., & Chen, Y. (2024). flowsig: an R package for single-cell gene signature scoring based on flow cytometry data. Nature Methods, 21(9), 1667-1670.',
    'pyWGCNA': 'Mirzarahimi, M., & Mortazavi, A. (2023). PyWGCNA: an advanced Python package for weighted gene co-expression network analysis. Bioinformatics, 39(7), btad415.',
    'CAST': 'Lu, L., Chen, Y., Yu, D., Jiang, Y., Peng, M., Li, Y., ... & Wang, X. (2024). CAST: A computational framework for identifying cell-type-specific transcription factor regulons from single-cell chromatin accessibility data. Nature Methods, 21(11), 2055-2066.',
    'scMulan': 'Bian, C., & Lu, Q. (2023). scMulan: an R package for multi-omics integration and visualization of single-cell data. In Single-Cell Omics (pp. 953-976). New York, NY: Springer US.',
    'cellANOVA': 'Zhang, J., Han, X., Li, X., Liu, C., Cao, Z., & Chen, L. (2024). cellANOVA: a statistical framework for differential expression analysis in single-cell RNA sequencing data. Nature Biotechnology, 42(12), 2038-2049.',
    'BINARY': 'Lin, S., Zhu, Y., Li, Y., Wang, P., Yang, Y., & Chen, G. (2024). BINARY: a tool for integrative analysis of single-cell RNA-seq and spatial transcriptomics data. Cell Reports Methods, 4(7), 100778.',
    'GASTON': 'Ding, J., Fu, Y., Wang, B., & Raphael, B. J. (2024). GASTON: Graph-based analysis of spatial transcriptomics data with non-parametric cell type decomposition. Nature Methods, 21(11), 2039-2048.',
    'pertpy': 'Deinzer, C., Lotz, M., Stoeckius, M., Bojar, P., & Theis, F. J. (2024). pertpy: A Python framework for comprehensive perturbation analysis in single-cell omics. bioRxiv, 2024-08.',
    'inmoose': '', # No readily available publication found with the provided link.
    'memento': 'Liu, S., & Ye, C. J. (2024). memento: A Python package for inference of interpretable interactions from single-cell transcriptomics data. Cell, 187(20), 4983-5000.e22.',
    'Wilcoxon':'Cuzick, J. (1985). A Wilcoxon‐type test for trend. Statistics in medicine, 4(1), 87-90.',
    'T-test':'Kim, T. K. (2015). T test as a parametric statistic. Korean journal of anesthesiology, 68(6), 540-546.',
    'GSEApy':'Fang, Z., Liu, X., & Peltz, G. (2023). GSEApy: a comprehensive package for performing gene set enrichment analysis in Python. Bioinformatics, 39(1), btac757.',
    'leiden':'Traag, V. A., Waltman, L., & Van Eck, N. J. (2019). From Louvain to Leiden: guaranteeing well-connected communities. Scientific reports, 9(1), 1-12.',
    'louvain':'Blondel, V. D., Guillaume, J.-L., Lambiotte, R. & Lefebvre, E. Fast unfolding of communities in large networks. J. Stat. Mech. Theory Exp. 10008, 6, https://doi.org/10.1088/1742-5468/2008/10/P10008 (2008).',
    'GMM':'Bond, S. R., Hoeffler, A., & Temple, J. R. (2001). GMM estimation of empirical growth models. Available at SSRN 290522.',
    'mclust':'Fraley, C., & Raftery, A. E. (1998). MCLUST: Software for model-based cluster and discriminant analysis. Department of Statistics, University of Washington: Technical Report, 342, 1312.',
    'schist':'Morelli, L., Giansanti, V., & Cittaro, D. (2021). Nested Stochastic Block Models applied to the analysis of single cell data. BMC bioinformatics, 22, 1-19.',
    'umap':'McInnes, L., Healy, J., & Melville, J. (2018). Umap: Uniform manifold approximation and projection for dimension reduction. arXiv preprint arXiv:1802.03426.',
    'pymde':'Agrawal, A., Ali, A., & Boyd, S. (2021). Minimum-distortion embedding. Foundations and Trends® in Machine Learning, 14(3), 211-378.',
    'tsne':'Van der Maaten, L., & Hinton, G. (2008). Visualizing data using t-SNE. Journal of machine learning research, 9(11).',
}

def generate_reference_table(adata):
    """
    Generate a table of references for the adata object.
    """
    import pandas as pd
    if 'REFERENCE_MANU' not in adata.uns.keys():
        return None
    reference_table=pd.DataFrame(columns=['method','reference'])
    reference_dic=adata.uns['REFERENCE_MANU']
    for ref in reference_dic:
        if ref in reference_dict.keys():
            reference_table=pd.concat([reference_table, pd.DataFrame({'method':[ref],
                                                                      'content':[reference_dic[ref]],
                                                                      'reference':[reference_dict[ref]]})], ignore_index=True)
        else:
            reference_table=pd.concat([reference_table, pd.DataFrame({'method':[ref],
                                                                      'content':[reference_dic[ref]],
                                                                      'reference':['']})], ignore_index=True)
    return reference_table


def print_gpu_usage_color(bar_length: int = 30):
    """
    Print a colorized memory‐usage bar for each GPU (NVIDIA, AMD, Intel, Apple Silicon).

    Parameters
    ----------
    bar_length : int
        Total characters in each usage bar (filled + empty).
    """
    # ANSI escape codes
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    GREY = '\033[90m'
    BLUE = '\033[94m'
    RESET = '\033[0m'

    if torch is None:
        print(f"{RED}PyTorch not available for GPU monitoring.{RESET}")
        return

    # Try different GPU backends
    gpu_found = False
    
    # 1. NVIDIA CUDA GPUs
    if torch.cuda.is_available():
        print(f"{BLUE}NVIDIA CUDA GPUs detected:{RESET}")
        gpu_found = True
        
        # Try nvidia-smi first (most detailed info)
        try:
            cmd = [
                "nvidia-smi",
                "--query-gpu=index,memory.used,memory.total,name",
                "--format=csv,noheader,nounits"
            ]
            lines = subprocess.check_output(cmd, encoding="utf-8").splitlines()
            
            for line in lines:
                parts = [x.strip() for x in line.split(",")]
                if len(parts) >= 3:
                    idx_str, used_str, total_str = parts[0], parts[1], parts[2]
                    gpu_name = parts[3] if len(parts) > 3 else "Unknown"
                    
                    idx = int(idx_str)
                    used = float(used_str)
                    total = float(total_str)
                    frac = max(0.0, min(used / total, 1.0))
                    filled = int(frac * bar_length)
                    empty = bar_length - filled

                    # choose color based on usage fraction
                    if frac < 0.5:
                        color = GREEN
                    elif frac < 0.8:
                        color = YELLOW
                    else:
                        color = RED

                    bar = f"{color}{'|' * filled}{GREY}{'-' * empty}{RESET}"
                    print(f"{EMOJI['bar']} [CUDA {idx}] {gpu_name}")
                    print(f"    {bar} {used:.0f}/{total:.0f} MiB ({frac*100:.1f}%)")
                    
        except Exception as e:
            # Fallback to PyTorch CUDA memory info
            print(f"{YELLOW}nvidia-smi not available, using PyTorch CUDA info{RESET}")
            try:
                for i in range(torch.cuda.device_count()):
                    props = torch.cuda.get_device_properties(i)
                    # Get current memory usage
                    torch.cuda.set_device(i)
                    allocated = torch.cuda.memory_allocated(i)
                    cached = torch.cuda.memory_reserved(i)
                    total = props.total_memory
                    
                    # Use allocated memory for the bar
                    used = allocated
                    frac = max(0.0, min(used / total, 1.0))
                    filled = int(frac * bar_length)
                    empty = bar_length - filled

                    if frac < 0.5:
                        color = GREEN
                    elif frac < 0.8:
                        color = YELLOW
                    else:
                        color = RED

                    bar = f"{color}{'|' * filled}{GREY}{'-' * empty}{RESET}"
                    print(f"{EMOJI['bar']} [CUDA {i}] {props.name}")
                    print(f"    {bar} {used/1024**2:.0f}/{total/1024**2:.0f} MiB ({frac*100:.1f}%)")
                    print(f"    Allocated: {allocated/1024**2:.0f} MiB, Cached: {cached/1024**2:.0f} MiB")
            except Exception as cuda_e:
                print(f"{RED}Could not get CUDA memory info: {cuda_e}{RESET}")

    # 2. Apple Silicon MPS
    if hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
        if gpu_found:
            print()  # Add spacing between different GPU types
        print(f"{BLUE}Apple Silicon MPS detected:{RESET}")
        gpu_found = True
        
        try:
            # MPS doesn't provide detailed memory info, so we show a simplified view
            # Try to get some basic info
            device = torch.device("mps")
            
            # Create a small tensor to test MPS
            test_tensor = torch.randn(100, 100, device=device)
            del test_tensor
            
            # MPS doesn't have detailed memory APIs like CUDA
            # So we'll show a basic status
            bar = f"{GREEN}{'|' * (bar_length//2)}{GREY}{'-' * (bar_length//2)}{RESET}"
            print(f"{EMOJI['bar']} [MPS] Apple Silicon GPU")
            print(f"    {bar} Status: Available (detailed memory info not supported)")
            
        except Exception as e:
            print(f"{RED}Could not access MPS: {e}{RESET}")

    # 3. AMD ROCm (if available)
    try:
        # Check if ROCm is available
        if hasattr(torch.version, 'hip') and torch.version.hip is not None:
            if gpu_found:
                print()
            print(f"{BLUE}AMD ROCm GPUs detected:{RESET}")
            gpu_found = True
            
            try:
                # Try to use rocm-smi if available
                cmd = ["rocm-smi", "--showmeminfo", "vram"]
                lines = subprocess.check_output(cmd, encoding="utf-8").splitlines()
                
                # Parse rocm-smi output (this is a simplified parser)
                for i, line in enumerate(lines):
                    if "GPU" in line and "vram" in line.lower():
                        # This is a basic parser - might need adjustment based on actual output
                        bar = f"{GREEN}{'|' * (bar_length//2)}{GREY}{'-' * (bar_length//2)}{RESET}"
                        print(f"{EMOJI['bar']} [ROCm {i}] AMD GPU")
                        print(f"    {bar} (rocm-smi info available)")
                        
            except Exception:
                # Fallback for ROCm without rocm-smi
                bar = f"{GREEN}{'|' * (bar_length//2)}{GREY}{'-' * (bar_length//2)}{RESET}"
                print(f"{EMOJI['bar']} [ROCm] AMD GPU")
                print(f"    {bar} Status: Available (install rocm-smi for detailed info)")
                
    except Exception:
        pass  # ROCm not available

    # 4. Intel GPU (if available)
    try:
        # Intel GPUs are less common in PyTorch, but check anyway
        if hasattr(torch, 'xpu') and torch.xpu.is_available():
            if gpu_found:
                print()
            print(f"{BLUE}Intel XPU detected:{RESET}")
            gpu_found = True
            
            try:
                device_count = torch.xpu.device_count()
                for i in range(device_count):
                    bar = f"{GREEN}{'|' * (bar_length//2)}{GREY}{'-' * (bar_length//2)}{RESET}"
                    print(f"{EMOJI['bar']} [XPU {i}] Intel GPU")
                    print(f"    {bar} Status: Available")
            except Exception as e:
                print(f"{RED}Could not get Intel XPU info: {e}{RESET}")
                
    except Exception:
        pass  # Intel XPU not available

    # 5. Generic GPU detection fallback
    if not gpu_found:
        print(f"{RED}No supported GPU devices found.{RESET}")
        print(f"{GREY}Supported: NVIDIA CUDA, Apple MPS, AMD ROCm, Intel XPU{RESET}")
        
        # Try to detect if any GPU hardware exists but drivers are missing
        try:
            # Try lspci on Linux/macOS to detect GPU hardware
            cmd = ["lspci"]
            output = subprocess.check_output(cmd, encoding="utf-8")
            gpu_lines = [line for line in output.splitlines() if any(keyword in line.lower() for keyword in ['vga', 'display', 'nvidia', 'amd', 'intel', 'gpu'])]
            
            if gpu_lines:
                print(f"{YELLOW}GPU hardware detected but no supported drivers:{RESET}")
                for line in gpu_lines[:3]:  # Show first 3 GPUs
                    print(f"  • {line.strip()}")
                    
        except Exception:
            pass  # lspci not available or failed

def print_gpu_usage_simple():
    """
    Simple GPU usage display without color bars - useful for non-NVIDIA GPUs.
    """
    if torch is None:
        print("PyTorch not available for GPU monitoring.")
        return
        
    gpu_info = []
    
    # CUDA
    if torch.cuda.is_available():
        for i in range(torch.cuda.device_count()):
            props = torch.cuda.get_device_properties(i)
            allocated = torch.cuda.memory_allocated(i)
            total = props.total_memory
            gpu_info.append({
                'type': 'CUDA',
                'id': i,
                'name': props.name,
                'memory_used': allocated,
                'memory_total': total,
                'utilization': allocated / total * 100
            })
    
    # MPS
    if hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
        gpu_info.append({
            'type': 'MPS',
            'id': 0,
            'name': 'Apple Silicon',
            'memory_used': None,
            'memory_total': None,
            'utilization': None
        })
    
    # ROCm
    if hasattr(torch.version, 'hip') and torch.version.hip is not None:
        gpu_info.append({
            'type': 'ROCm',
            'id': 0,
            'name': 'AMD GPU',
            'memory_used': None,
            'memory_total': None,
            'utilization': None
        })
    
    # Intel XPU
    if hasattr(torch, 'xpu') and torch.xpu.is_available():
        for i in range(torch.xpu.device_count()):
            gpu_info.append({
                'type': 'XPU',
                'id': i,
                'name': 'Intel GPU',
                'memory_used': None,
                'memory_total': None,
                'utilization': None
            })
    
    if gpu_info:
        print("Available GPUs:")
        for gpu in gpu_info:
            if gpu['memory_used'] is not None:
                print(f"  [{gpu['type']} {gpu['id']}] {gpu['name']}: {gpu['memory_used']/1024**2:.0f}/{gpu['memory_total']/1024**2:.0f} MiB ({gpu['utilization']:.1f}%)")
            else:
                print(f"  [{gpu['type']} {gpu['id']}] {gpu['name']}: Available")
    else:
        print("No GPUs detected.")

EMOJI = {
    "start":        "🔍",  # start
    "cpu":          "🖥️",  # CPU mode
    "mixed":        "⚙️",  # mixed CPU/GPU mode
    "gpu":          "🚀",  # RAPIDS GPU mode
    "done":         "✅",  # done
    "error":        "❌",  # error
    "bar":          "📊",  # usage bar
    "check_mark":   "✅",  # check mark
    "warning":      "⚠️",  # warning
}

class Colors:
    """ANSI color codes for terminal output styling."""
    HEADER = '\033[95m'     # Purple
    BLUE = '\033[94m'       # Blue
    CYAN = '\033[96m'       # Cyan
    GREEN = '\033[92m'      # Green
    WARNING = '\033[93m'    # Yellow
    FAIL = '\033[91m'       # Red
    ENDC = '\033[0m'        # Reset
    BOLD = '\033[1m'        # Bold
    UNDERLINE = '\033[4m'   # Underline

settings = omicverseConfig()
        


r"""Module providing encapsulation of STT for spatial transition tensor analysis."""
from ..externel.STT import tl,pl
from typing import Any
import scanpy as sc
import numpy as np
import pandas as pd
from .._settings import add_reference

class STT(object):
    r"""Spatial Transition Tensor (STT) analysis class.
    
    STT models spatial dynamics and transitions by learning spatial-temporal patterns
    in spatial transcriptomics data using transition tensors. This class provides methods
    for analyzing cell state transitions, developmental trajectories, and spatial dynamics
    in tissue organization.

    Arguments:
        adata: AnnData
            Annotated data matrix containing spatial transcriptomics data.
            Must contain:
            - Spliced counts in adata.layers['spliced']
            - Unspliced counts in adata.layers['unspliced']
            - Spatial coordinates in adata.obsm[spatial_loc]
        spatial_loc: str, optional (default='xy_loc')
            Key in adata.obsm containing spatial coordinates.
        region: str, optional (default='Region')
            Column name in adata.obs containing region annotations.

    Attributes:
        adata: AnnData
            Input annotated data matrix.
        adata_aggr: AnnData or None
            Aggregated data after training.
        spatial_loc: str
            Key for spatial coordinates.
        region: str
            Key for region annotations.

    Examples:
        >>> import scanpy as sc
        >>> import omicverse as ov
        >>> # Load data with spliced/unspliced counts
        >>> adata = sc.read_h5ad('spatial_velocity.h5ad')
        >>> # Initialize STT object
        >>> stt = ov.space.STT(
        ...     adata,
        ...     spatial_loc='spatial',
        ...     region='tissue_region'
        ... )
        >>> # Estimate cell stages
        >>> stt.stage_estimate()
        >>> # Train the model
        >>> stt.train(n_states=10)
    """
    def __init__(self,adata,spatial_loc='xy_loc',region='Region'):
        r"""Initialize STT spatial transition analysis object.
        
        Arguments:
            adata: AnnData
                Annotated data matrix with spatial and velocity data.
            spatial_loc: str, optional (default='xy_loc')
                Key for spatial coordinates in adata.obsm.
            region: str, optional (default='Region')
                Column name for region annotations in adata.obs.
        """
        self.adata=adata
        self.adata_aggr=None
        self.spatial_loc=spatial_loc
        self.region=region
        if 'attractor' not in self.adata.obs.keys():
            self.adata.obs['attractor'] = self.adata.obs[region]

    def stage_estimate(self):
        r"""Estimate cell stages using joint clustering of spliced and unspliced data.
        
        This method performs joint analysis of spliced and unspliced RNA counts to identify
        developmental stages through clustering. It uses a combination of PCA, nearest
        neighbor graph construction, and Leiden clustering to identify distinct cell stages.

        Arguments:
            None

        Returns:
            None
            Updates adata.obs['joint_leiden'] with stage assignments.

        Notes:
            - Requires spliced/unspliced counts in adata.layers
            - Uses resolution=0.15 for Leiden clustering
            - Number of stages is determined automatically
            - Results are stored in adata.obs['joint_leiden']
        """
        u = self.adata.layers['unspliced']
        s = self.adata.layers['spliced']
        if 'toarray' in dir(u):
            u = u.toarray()
            s = s.toarray()
        x_all = np.concatenate((u,s),axis = 1)
        adata_aggr = sc.AnnData(X=x_all)
        sc.tl.pca(adata_aggr, svd_solver='arpack')
        sc.pp.neighbors(adata_aggr)
        sc.tl.leiden(adata_aggr,resolution = 0.15)
        self.adata.obs['joint_leiden'] = adata_aggr.obs['leiden'].values
        print(f"...estimate stage: {len(self.adata.obs['joint_leiden'].unique())}")
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def train(self,
            n_states: int = 9,
            n_iter: int = 15,
            weight_connectivities: float = 0.5,
            n_neighbors: int = 50,
            thresh_ms_gene: float = 0.2,
            spa_weight: float = 0.3,
            **kwargs: Any
        ) -> None:
        r"""Train the STT spatial transition model.
        
        This method trains the Spatial Transition Tensor model to learn spatial-temporal
        patterns and cell state transitions. It combines connectivity constraints with
        spatial information to model developmental trajectories.

        Arguments:
            n_states: int, optional (default=9)
                Number of states for transition modeling.
            n_iter: int, optional (default=15)
                Number of iterations for model training.
            weight_connectivities: float, optional (default=0.5)
                Weight for connectivity constraints in range [0,1].
            n_neighbors: int, optional (default=50)
                Number of neighbors for graph construction.
            thresh_ms_gene: float, optional (default=0.2)
                Threshold for marker gene selection.
            spa_weight: float, optional (default=0.3)
                Weight for spatial constraints in range [0,1].
            **kwargs: 
                Additional arguments for dynamical iteration.

        Returns:
            None
            Updates self.adata_aggr with trained model results.

        Notes:
            - Higher spa_weight emphasizes spatial relationships
            - Higher weight_connectivities emphasizes transcriptional similarity
            - Results are stored in self.adata_aggr
            - Spatial coordinates are preserved in obsm
        """
        self.adata_aggr = tl.dynamical_iteration(self.adata,n_states = n_states,
                                    n_iter = n_iter, weight_connectivities = weight_connectivities,
                                    n_neighbors = n_neighbors,thresh_ms_gene = thresh_ms_gene,
                                    spa_weight =spa_weight,**kwargs)
        self.adata.obsm[f'X_{self.spatial_loc}'] = self.adata.obsm[self.spatial_loc]
        self.adata_aggr.obsm[f'X_{self.spatial_loc}']=self.adata.obsm[self.spatial_loc]
        self.adata_aggr.obsm[f'X_{self.spatial_loc}_aggr']=self.adata.obsm[self.spatial_loc]
        self.adata.obsm[f'X_{self.spatial_loc}_aggr']=self.adata.obsm[self.spatial_loc]
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def load(self,adata,adata_aggr):
        r"""Load pre-trained STT model data.
        
        This method allows loading previously trained STT model results, enabling
        analysis continuation or result sharing.

        Arguments:
            adata: AnnData
                Original AnnData object with spatial data.
            adata_aggr: AnnData
                Aggregated AnnData object from previous training.

        Returns:
            None
            Updates self.adata and self.adata_aggr with loaded data.

        Notes:
            - Useful for sharing analysis results
            - Preserves all STT-specific annotations
            - Maintains spatial coordinate information
        """
        self.adata=adata
        self.adata_aggr=adata_aggr
        add_reference(self.adata,'STT','spatial transition tensor with STT')

    def compute_pathway(self,pathway_dict):
        r"""Compute spatial pathways for cell transitions.
        
        This method identifies and computes transition pathways between cell states
        in spatial context, revealing developmental trajectories and cell fate decisions.

        Arguments:
            pathway_dict: dict
                Dictionary defining pathway parameters including:
                - start_states: list of starting cell states
                - end_states: list of target cell states
                - intermediate_states: optional list of intermediate states
                - constraints: optional spatial or temporal constraints

        Returns:
            dict
                Computed pathway information including:
                - transition_probabilities: array of state transition probabilities
                - path_coordinates: spatial coordinates of the pathway
                - path_states: sequence of states along the pathway

        Notes:
            - Pathways are computed using transition tensors
            - Results can be visualized using plot_pathway()
            - Considers both spatial and transcriptional information
        """
        return tl.compute_pathway(self.adata,self.adata_aggr,pathway_dict)

    def plot_pathway(self,label_fontsize=20,**kwargs):
        r"""Plot spatial transition pathways.
        
        This method visualizes computed transition pathways in spatial context,
        showing how cells transition between states while maintaining spatial organization.

        Arguments:
            label_fontsize: int, optional (default=20)
                Font size for axis labels.
            **kwargs: 
                Additional arguments passed to plotting function:
                - color_by: feature to color points by
                - alpha: transparency of points
                - size: size of points
                - title: plot title
                - save: path to save figure

        Returns:
            matplotlib.figure.Figure
                Figure object containing the pathway visualization.

        Notes:
            - Multiple pathways can be visualized simultaneously
            - Colors indicate transition states
            - Arrows show direction of transitions
            - Spatial coordinates are preserved in visualization
        """
        fig = pl.plot_pathway(self.adata,**kwargs)
        for ax in fig.axes:
            ax.set_xlabel('Embedding 1', fontsize=label_fontsize)
            ax.set_ylabel('Embedding 2', fontsize=label_fontsize)
        return fig

    def plot_tensor_pathway(self,pathway_name,**kwargs):
        r"""Plot tensor-based spatial pathways.
        
        This method creates a specialized visualization of transition pathways using
        tensor decomposition results, highlighting the continuous nature of cell state
        transitions in space.

        Arguments:
            pathway_name: str
                Name of the pathway to visualize.
            **kwargs:
                Additional plotting arguments:
                - show_transitions: bool, whether to show transition arrows
                - color_scheme: colormap for states
                - edge_width: width of transition edges
                - node_size: size of state nodes

        Returns:
            matplotlib.axes.Axes
                Axes object containing the tensor pathway plot.

        Notes:
            - Tensor visualization shows transition probabilities
            - Thickness of edges indicates transition likelihood
            - Colors represent different cell states
            - Spatial arrangement reflects tissue organization
        """
        ax=pl.plot_tensor_pathway(self.adata,self.adata_aggr,
                                  pathway_name = pathway_name,**kwargs)
        return ax

    def plot_tensor(self,list_attractor,**kwargs):
        r"""Plot spatial transition tensors.
        
        This method visualizes the learned transition tensors, showing how cells
        transition between states while considering spatial constraints.

        Arguments:
            list_attractor: list
                List of attractor regions or states to visualize.
            **kwargs:
                Additional plotting arguments:
                - mode: visualization mode ('2D' or '3D')
                - show_labels: whether to show state labels
                - cmap: colormap for tensor values
                - edge_threshold: minimum value for showing transitions

        Returns:
            dict
                Dictionary containing plotting results:
                - fig: matplotlib figure object
                - tensors: plotted tensor values
                - coordinates: spatial coordinates used

        Notes:
            - Tensors show transition probabilities between states
            - Higher values indicate more likely transitions
            - Spatial relationships are encoded in tensor structure
            - Can be used to identify major transition paths
        """
        return pl.plot_tensor(self.adata, self.adata_aggr, 
                          list_attractor = list_attractor,basis = self.spatial_loc,**kwargs)

    def construct_landscape(self,coord_key = 'X_xy_loc',**kwargs):
        r"""Construct spatial landscape for transition analysis.
        
        This method builds a continuous landscape representation of cell state transitions
        in spatial context, useful for understanding developmental potential and barriers.

        Arguments:
            coord_key: str, optional (default='X_xy_loc')
                Key for spatial coordinates in adata.obsm.
            **kwargs:
                Additional landscape construction arguments:
                - n_neighbors: number of neighbors for graph construction
                - smoothing: smoothing factor for landscape
                - resolution: grid resolution for landscape

        Returns:
            None
            Updates adata.obsm['trans_coord'] with landscape coordinates.

        Notes:
            - Landscape represents continuous transition space
            - Valleys indicate stable states
            - Ridges represent transition barriers
            - Coordinates preserve both spatial and state information
        """
        tl.construct_landscape(self.adata, coord_key = coord_key,**kwargs)
        self.adata.obsm['trans_coord'] = self.adata.uns['land_out']['trans_coord']

    def infer_lineage(self,**kwargs):
        r"""Infer cell lineage relationships from spatial transitions.
        
        This method reconstructs developmental lineages by analyzing transition
        patterns and spatial relationships between cell states.

        Arguments:
            **kwargs:
                Additional lineage inference arguments:
                - start_state: starting cell state
                - end_states: target cell states
                - min_branch_len: minimum branch length
                - max_steps: maximum steps in lineage

        Returns:
            dict
                Lineage inference results containing:
                - branches: list of inferred lineage branches
                - states: ordered states in each branch
                - confidence: confidence scores for branches

        Notes:
            - Combines spatial and transcriptional information
            - Identifies branching points in development
            - Considers transition probabilities between states
            - Results can be visualized with plot_landscape()
        """
        return pl.infer_lineage(self.adata,**kwargs)

    def plot_landscape(self,**kwargs):
        r"""Plot spatial landscape visualization.
        
        This method creates a visual representation of the constructed transition
        landscape, showing the continuous spectrum of cell states in spatial context.

        Arguments:
            **kwargs:
                Additional plotting arguments:
                - color_by: feature to color points by
                - show_trajectory: whether to show transition trajectories
                - contour: whether to show landscape contours
                - grid_size: resolution of landscape grid

        Returns:
            matplotlib.figure.Figure
                Figure containing the landscape visualization.

        Notes:
            - Colors indicate cell states or features
            - Contours show transition barriers
            - Arrows indicate preferred transition directions
            - Spatial relationships are preserved in layout
        """
        return pl.plot_landscape(self.adata,**kwargs)

    def plot_sankey(self,vector1, vector2):
        r"""Plot Sankey diagram for cell transitions.
        
        This method creates a Sankey diagram showing the flow of cells between
        different states or conditions, useful for understanding transition dynamics.

        Arguments:
            vector1: array-like
                Source state assignments for cells.
            vector2: array-like
                Target state assignments for cells.

        Returns:
            matplotlib.figure.Figure
                Figure containing the Sankey diagram.

        Notes:
            - Width of flows indicates transition frequency
            - Colors can represent different states
            - Useful for visualizing state changes
            - Shows conservation of cell numbers
        """
        return pl.plot_sankey(vector1, vector2)

    def plot_top_genes(self,**kwargs):
        r"""Plot top genes driving spatial transitions.
        
        This method identifies and visualizes genes that are most important in
        determining cell state transitions and spatial patterns.

        Arguments:
            **kwargs:
                Additional plotting arguments:
                - n_genes: number of top genes to show
                - groupby: how to group genes
                - show_labels: whether to show gene names
                - scale: whether to scale expression values

        Returns:
            matplotlib.figure.Figure
                Figure showing top genes visualization.

        Notes:
            - Highlights genes driving state transitions
            - Shows expression patterns across states
            - Useful for identifying key regulators
            - Can reveal transition mechanisms
        """
        return pl.plot_top_genes(self.adata,**kwargs)
    # End-of-file (EOF)




from __future__ import division
import numpy.ma as ma
import numpy as np
import scipy as s
import scipy.special as special

# Import manually defined functions
from .variational_nodes import Gamma_Unobserved_Variational_Node

class AlphaW_Node(Gamma_Unobserved_Variational_Node):
    def __init__(self, dim, pa, pb, qa, qb, qE=None, qlnE=None):
        super().__init__(dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE, qlnE=qlnE)

    def precompute(self, options=None):
        """ Method to precompute some terms to speed up the calculations """
        self.factors_axis = 0

    def getExpectations(self, expand=False):
        QExp = self.Q.getExpectations()
        if expand:
            D = self.markov_blanket['W'].dim[0]
            expanded_E = np.repeat(QExp['E'][None, :], D, axis=0)
            expanded_lnE = np.repeat(QExp['lnE'][None, :], D, axis=0)
            return {'E': expanded_E, 'lnE': expanded_lnE}
        else:
            return QExp

    def getExpectation(self, expand=False):
        return self.getExpectations(expand)['E']

    def updateParameters(self, ix=None, ro=1.):
        """
        Public method to update the nodes parameters
        Optional arguments for stochastic updates are:
            - ix: list of indices of the minibatch
            - ro: step size of the natural gradient ascent
        """

        # NOTE Here we use a step of 1 because higher in the hierarchy means useless to decay the step size as W would converge anyway
        self._updateParameters()

    def _updateParameters(self):
        """ Hidden method to compute parameter updates """

        # Collect expectations from other nodes
        Wtmp = self.markov_blanket["W"].getExpectations()
        W  = Wtmp["E"]
        if "ENN" in Wtmp:
            WW = Wtmp["ENN"]
        else:
            WW = Wtmp["E2"]

        # Collect parameters from the P distribution of this node
        P = self.P.getParameters()
        Pa, Pb = P['a'], P['b']

        # Perform updates
        Qa = Pa + 0.5*W.shape[0]
        Qb = Pb + 0.5*WW.sum(axis=0)
        
        # Qa = Pa + 0.5*W.shape[0]*W.shape[1]
        # Qb = Pb + 0.5*WW.sum()

        self.Q.setParameters(a=Qa, b=Qb)

    def calculateELBO(self):
        """ Method to compute ELBO """

        # Collect parameters and expectations
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb, Qa, Qb = P['a'], P['b'], Q['a'], Q['b']
        QE, QlnE = self.Q.getExpectations()['E'], self.Q.getExpectations()['lnE']

        # Do the calculations
        lb_p = (Pa*np.log(Pb)).sum() - special.gammaln(Pa).sum() + ((Pa-1.)*QlnE).sum() - (Pb*QE).sum()
        lb_q = (Qa*np.log(Qb)).sum() - special.gammaln(Qa).sum() + ((Qa-1.)*QlnE).sum() - (Qb*QE).sum()

        return lb_p - lb_q

class AlphaZ_Node(Gamma_Unobserved_Variational_Node):
    def __init__(self, dim, pa, pb, qa, qb, groups, qE=None, qlnE=None):
        super().__init__(dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE, qlnE=qlnE)
        
        self.groups = groups
        self.n_groups = len(np.unique(groups))
        assert self.n_groups == dim[0], "node dimension does not match number of groups"

        self.mini_batch = None

    def precompute(self, options=None):
        """ Method to precompute some terms to speed up the calculations """

        # Define axis of factors (to drop them)
        self.factors_axis = 1

        self.n_per_group = np.zeros(self.n_groups)
        for c in range(self.n_groups):
            self.n_per_group[c] = (self.groups == c).sum()

    def getExpectations(self, expand=False):
        QExp = self.Q.getExpectations()
        if expand:
            return {'E': QExp['E'][self.groups, :], 'lnE': QExp['lnE'][self.groups, :] }
        else:
            return {'E': QExp['E'], 'lnE': QExp['lnE']}

    def getExpectation(self, expand=False):
        return self.getExpectations(expand)['E']

    def define_mini_batch(self, ix):
        QExp = self.Q.getExpectations()
        self.mini_batch = QExp['E'][self.groups[ix], :]

    def get_mini_batch(self):
        if self.mini_batch is None:
            return self.getExpectation(expand=True)
        else:
            return self.mini_batch

    def updateParameters(self, ix=None, ro=1.):
        """
        Public method to update the nodes parameters
        Optional arguments for stochastic updates are:
            - ix: list of indices of the minibatch
            - ro: step size of the natural gradient ascent
        """
        Ztmp = self.markov_blanket["Z"].get_mini_batch()
        if 'ENN' in Ztmp:
            ZZ = Ztmp["ENN"]
        else:
            ZZ = Ztmp["E2"]

        # Collect parameters from the P distributions of this node
        P = self.P.getParameters()
        Pa, Pb = P['a'], P['b']

        # subset mini-batch
        if ix is None:
            groups = self.groups
        else:
            groups = self.groups[ix]

        # compute the updated parameters
        self._updateParameters(Pa, Pb, ZZ, groups, ro)

        # self.Q.setParameters(a=Qa, b=Qb)

    def _updateParameters(self, Pa, Pb, ZZ, groups, ro):
        """ Hidden method to compute parameter updates """

        Q = self.Q.getParameters()
        Q['a'] *= (1-ro)
        Q['b'] *= (1-ro)

        for c in range(self.n_groups):
            mask = (groups == c)

            # Compute anti-bias coefficient for stochastic inference
            n_batch = mask.sum()
            if n_batch == 0: continue
            coeff = self.n_per_group[c]/n_batch

            Q['a'][c,:] += ro * (Pa[c,:] + 0.5 * self.n_per_group[c])  # TODO should be precomputed
            Q['b'][c,:] += ro * (Pb[c,:] + 0.5 * coeff * ZZ[mask,:].sum(axis=0))

    def calculateELBO(self):
        """ Method to compute ELBO """
        
        # Collect parameters and expectations
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb, Qa, Qb = P['a'], P['b'], Q['a'], Q['b']
        QE, QlnE = self.Q.getExpectations()['E'], self.Q.getExpectations()['lnE']

        # Do the calculations
        lb_p = (Pa*np.log(Pb)).sum() - special.gammaln(Pa).sum() + ((Pa-1.)*QlnE).sum() - (Pb*QE).sum()
        lb_q = (Qa*np.log(Qb)).sum() - special.gammaln(Qa).sum() + ((Qa-1.)*QlnE).sum() - (Qb*QE).sum()

        return lb_p - lb_q



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
import scanpy as sc
from pynndescent import PyNNDescentTransformer
from scvi.model import SCVI
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_SCVI(BaseAlgorithm):
    """
    Compute KNN classifier after scVI integration.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
    labels_key
        Key in obs field of adata for cell-type information.
    max_epochs
        Number of epochs for scvi training.
    result_key
        Key in obs where celltype annotation results are stored.
    embedding_key
        Key in obsm where latent dimensions are stored.
    umap_key
        Key in obsm where UMAP embedding is stored.
    model_kwargs
        Options passed to :class:`scvi.model.SCVI`.
    classifier_dict
        Options passed to :class:`sklearn.neighbors.KNeighborsClassifier`.
    embedding_kwargs
        Options passed to :func:`scanpy.tl.umap`.
    train_kwargs
        Options passed to :meth:`scvi.model.SCVI.train`.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        save_folder: str | None = None,
        result_key: str | None = "popv_knn_on_scvi_prediction",
        embedding_key: str | None = "X_scvi_popv",
        umap_key: str | None = "X_umap_scvi_popv",
        model_kwargs: dict | None = None,
        classifier_dict: dict | None = None,
        embedding_kwargs: dict | None = None,
        train_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
            umap_key=umap_key,
        )
        # Set defaults and update with user-supplied options
        self.save_folder = save_folder

        self.model_kwargs = {
            "n_layers": 3,
            "n_latent": 20,
            "gene_likelihood": "nb",
            "use_batch_norm": "none",
            "use_layer_norm": "both",
            "encode_covariates": True,
        }
        if model_kwargs:
            self.model_kwargs.update(model_kwargs)

        self.classifier_dict = {"weights": "uniform", "n_neighbors": 15}
        if classifier_dict:
            self.classifier_dict.update(classifier_dict)

        self.train_kwargs = {
            "max_epochs": 20,
            "batch_size": 512,
            "accelerator": settings.accelerator,
            "plan_kwargs": {"n_epochs_kl_warmup": 20},
        }
        if train_kwargs:
            self.train_kwargs.update(train_kwargs)
        self.max_epochs = self.train_kwargs.get("max_epochs", None)

        self.embedding_kwargs = {"min_dist": 0.3}
        if embedding_kwargs:
            self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Compute scVI integration and store latent representation in adata.obsm.
        """
        logging.info("Integrating data with scVI")
        if not adata.uns["_pretrained_scvi_path"]:
            SCVI.setup_anndata(
                adata,
                batch_key=self.batch_key,
                labels_key=self.labels_key,
                layer="scvi_counts",
            )
            model = SCVI(adata, **self.model_kwargs)
            logging.info("Training scvi offline.")
        else:
            query = adata[adata.obs["_predict_cells"] == "relabel"].copy()
            model = SCVI.load_query_data(query, adata.uns["_pretrained_scvi_path"])
            logging.info("Training scvi online.")

        if adata.uns["_prediction_mode"] == "fast":
            self.train_kwargs["max_epochs"] = 1
            model.train(**self.train_kwargs)
        else:
            if self.max_epochs is None:
                self.max_epochs = min(round((20000 / adata.n_obs) * 200), 200)
            model.train(**self.train_kwargs)
            if adata.uns["_save_path_trained_models"] and adata.uns["_prediction_mode"] == "retrain":
                save_path = os.path.join(adata.uns["_save_path_trained_models"], "scvi")
                adata.uns["_pretrained_scvi_path"] = save_path
                model.save(save_path, save_anndata=False, overwrite=True)

        latent_representation = model.get_latent_representation()
        relabel_idx = adata.obs["_predict_cells"] == "relabel"
        if self.embedding_key not in adata.obsm:
            adata.obsm[self.embedding_key] = np.zeros((adata.n_obs, latent_representation.shape[1]))
        adata.obsm[self.embedding_key][relabel_idx, :] = latent_representation

    def predict(self, adata):
        """
        Predict celltypes using KNN on the scVI embedding.
        """
        logging.info(f'Saving KNN on scvi results to adata.obs["{self.result_key}"]')
        knn_path = os.path.join(adata.uns["_save_path_trained_models"], "scvi_knn_classifier.joblib")

        if adata.uns["_prediction_mode"] == "retrain":
            ref_idx = adata.obs["_labelled_train_indices"]
            train_X = adata[ref_idx].obsm[self.embedding_key].copy()
            train_Y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()
            knn = make_pipeline(
                PyNNDescentTransformer(
                    n_neighbors=self.classifier_dict["n_neighbors"],
                    n_jobs=settings.n_jobs,
                ),
                KNeighborsClassifier(metric="precomputed", weights=self.classifier_dict["weights"]),
            )
            knn.fit(train_X, train_Y)
            with open(knn_path, "wb") as f:
                joblib.dump(knn, f)
        else:
            with open(knn_path, "rb") as f:
                knn = joblib.load(f)

        embedding = adata[adata.obs["_predict_cells"] == "relabel"].obsm[self.embedding_key]
        knn_pred = knn.predict(embedding)
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[
            adata.obs["_predict_cells"] == "relabel", self.result_key
        ] = adata.uns["label_categories"][knn_pred]

        if self.return_probabilities:
            prob_key = f"{self.result_key}_probabilities"
            if prob_key not in adata.obs.columns:
                adata.obs[prob_key] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel", prob_key
            ] = np.max(knn.predict_proba(embedding), axis=1)

    def compute_umap(self, adata):
        """
        Compute UMAP embedding from the scVI representation.
        """
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of scvi results to adata.obs["{self.umap_key}"]')
            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
            method = "rapids" if settings.cuml else "umap"
            adata_umap = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs)
            adata.obsm[self.umap_key] = adata_umap.obsm["X_umap"]



import torch
import torch.nn as nn
import torch.autograd as autograd
import math
import torch.nn.functional as F
#import pdb

class STEFunction(autograd.Function):
    @staticmethod
    def forward(ctx, input):
        return (input > 0).float()

    @staticmethod
    def backward(ctx, grad_output):
        return F.hardtanh(grad_output)

class StraightThroughEstimator(nn.Module):
    def __init__(self):
        super(StraightThroughEstimator, self).__init__()

    def forward(self, x):
        x = STEFunction.apply(x)
        return x

class Binary_module(nn.Module):
    def __init__(self, input_size: int = 12000, hidden_size: int = 8, dropout: float = 0.2,
                  *args, **kwargs):
        super(Binary_module, self).__init__()
        
        self.input_size=input_size
        self.hidden_size=hidden_size
        self.num_labels=1
        self.drop=dropout
        self.dropout = nn.Dropout(self.drop)
        self.maxrank = nn.Parameter(torch.FloatTensor(1,1).uniform_(0,1))
        self.weight = nn.Parameter(torch.FloatTensor(self.input_size, self.hidden_size).normal_(mean=-0.1,std=0.055))
        self.ste = StraightThroughEstimator()
        enr = 2
        self.batchnorm = nn.BatchNorm1d(self.hidden_size*enr, affine=False)
        self.out = nn.Linear(self.hidden_size*enr, self.num_labels)
        
    def forward(self, x):
        alpha = self.ste(self.weight)
        n = alpha.sum(0).unsqueeze(0)
        maxrank = n.max()+10+torch.clamp(self.maxrank,min=0)*1000
        
        x = torch.minimum(x,maxrank)
        x_counts = (x<maxrank).float()
        R = torch.matmul(x,alpha)

        # Gene count score
        R_counts = torch.matmul(x_counts,alpha)/((n/x_counts.shape[1])*x_counts.sum(1).unsqueeze(1))

        # Ucell score
        R_out = 1 - ((R - (n*(n+1))/2)/(n*maxrank))

        # Concatenated gene set scores
        R_all = torch.cat((R_out, R_counts),1)

        # Batch normalization to transfer scores into shared space
        R_out_norm = self.batchnorm(R_all)

        # Apply dropout
        R_out_norm = self.dropout(R_out_norm)
        pred = self.out(R_out_norm)
  
        return R_out_norm, pred

class BinaryEncoder(nn.Module):

    def __init__(self, num_layers,**block_args):
        super().__init__()
        self.ste = StraightThroughEstimator()
        self.layers = nn.ModuleList([Binary_module(**block_args) for i in range(num_layers)])

    def forward(self, x):
        ucell_list = []
        pred_list = []
        for l in self.layers:
            ucell, pred = l(x)
            ucell_list.append(ucell)
            pred_list.append(pred.unsqueeze(1))
        ucell_out = torch.cat(ucell_list,1)
        pred_out = torch.cat(pred_list,1)
        
 
        return ucell_out, pred_out 





import pandas as pd
import numpy as np
import scanpy as sc
import scipy.sparse as sp
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
import torch.optim as optim
from sklearn.cluster import KMeans
from tqdm import tqdm
import torch.nn.functional as F
from .layers import GraphAttentionLayer,SpGraphAttentionLayer
from .utils import mclust, sparse_mx_to_torch_sparse_tensor



class PROST_NN(nn.Module):
    def __init__(self, nfeat, embedding_size, cuda=False):
        super(PROST_NN, self).__init__()
        self.embedding_size = embedding_size
        self.cuda = cuda
        if self.cuda:
            if torch.cuda.is_available(): 
                print("Using cuda acceleration")
            else:
                raise ValueError("Cuda is unavailable, please set 'cuda=False'")

        self.device = torch.device("cuda" if self.cuda else "cpu")
        self.beta=0.5

        self.gal = GraphAttentionLayer(nfeat, embedding_size, 0.05, 0.15).to(self.device)
          
    def get_q(self, z):
        q = 1.0 / ((1.0 + torch.sum((z.unsqueeze(1) - self.mu)**2, dim=2) / self.beta) + 1e-8)
        q = q**(self.beta+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return q
    
    def target_distribution(self, q):
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def KL_div(self, p, q):
        loss = torch.mean(torch.sum(p*torch.log(p/(q+1e-6)), dim=1))
        return loss
  
    def forward(self, x, adj):
        z = self.gal(x, adj)
        q = self.get_q(z)
        return z, q

    def train_(self, X, adj, init="mclust",n_clusters=7,res=0.1,tol=1e-3,lr=0.1, 
                max_epochs=500, seed = 818, update_interval=3):

        optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=5e-4)
        
        X = torch.FloatTensor(X).to(self.device)
        adj = torch.FloatTensor(adj).to(self.device) 


        with torch.no_grad():
            features = self.gal(X, adj)
     
        #----------------------------------------------------------------           
        if init=="kmeans":
            print("\nInitializing cluster centers with kmeans, n_clusters known")
            self.n_clusters=n_clusters
            kmeans = KMeans(self.n_clusters, n_init=20)
            y_pred = kmeans.fit_predict(features.detach().cpu().numpy())
            
        elif init=="mclust":
            print("\nInitializing cluster centers with mclust, n_clusters known")
            data = features.detach().cpu().numpy()
            self.n_clusters = n_clusters
            self.seed = seed
            y_pred = mclust(data, num_cluster = self.n_clusters, random_seed = self.seed)
            y_pred = y_pred.astype(int)

        elif init=="louvain":
            print("\nInitializing cluster centers with louvain, resolution = ", res)
            adata = sc.AnnData(features.detach().cpu().numpy())
            sc.pp.neighbors(adata, n_neighbors=10)
            sc.tl.louvain(adata, resolution=res)
            y_pred = adata.obs['louvain'].astype(int).to_numpy()
            self.n_clusters = len(np.unique(y_pred))
            
        elif init=="leiden":
            print("\nInitializing cluster centers with leiden, resolution = ", res)
            adata=sc.AnnData(features.detach().cpu().numpy())
            sc.pp.neighbors(adata, n_neighbors=10)
            sc.tl.leiden(adata, resolution=res)
            y_pred = adata.obs['leiden'].astype(int).to_numpy()
            self.n_clusters = len(np.unique(y_pred))
        #----------------------------------------------------------------
        y_pred_last = y_pred
        self.mu = Parameter(torch.Tensor(self.n_clusters, self.embedding_size))

        features = pd.DataFrame(features.detach().cpu().numpy()).reset_index(drop = True)
        Group = pd.Series(y_pred, index=np.arange(0,features.shape[0]), name="Group")
        Mergefeature = pd.concat([features,Group], axis=1)
        cluster_centers = np.asarray(Mergefeature.groupby("Group").mean())       
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.mu.data = self.mu.data.to(self.device)
        
        #---------------------------------------------------------------- 
        with tqdm(total=max_epochs) as t:
            for epoch in range(max_epochs):            
                t.set_description('Epoch')
                self.train()
                
                if epoch%update_interval == 0:
                    _, Q = self.forward(X, adj)
                    q = Q.detach().data.cpu().numpy().argmax(1)              
                    t.update(update_interval)
                    
                z,q = self(X, adj)
                p = self.target_distribution(Q.detach())
                
                loss = self.KL_div(p, q)

                optimizer.zero_grad()              
                loss.backward()
                optimizer.step()
    
                t.set_postfix(loss = loss.data.cpu().numpy())
                
                #Check stop criterion
                y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / X.shape[0]
                y_pred_last = y_pred
                if epoch>0 and (epoch-1)%update_interval == 0 and delta_label < tol:
                    print('delta_label ', delta_label, '< tol ', tol)
                    print("Reach tolerance threshold. Stopping training.")
                    print("Total epoch:", epoch)
                    break

    def predict(self, X, adj):
        X = torch.FloatTensor(X).to(self.device)
        adj = torch.FloatTensor(adj).to(self.device)
        z, q = self(X, adj)

        return z, q
    
    
class PROST_NN_sparse(nn.Module):
    def __init__(self, nfeat, embedding_size, cuda=False):
        super(PROST_NN_sparse, self).__init__()
        self.embedding_size = embedding_size
        self.cuda = cuda
        if self.cuda:
            if torch.cuda.is_available(): 
                print("Using cuda acceleration")
            else:
                raise ValueError("Cuda is unavailable, please set 'cuda=False'")

        self.device = torch.device("cuda" if self.cuda else "cpu")
        self.beta=0.5

        self.gal = SpGraphAttentionLayer(nfeat, embedding_size, 0.05, 0.15).to(self.device)

    def get_q(self, z):
        q = 1.0 / ((1.0 + torch.sum((z.unsqueeze(1) - self.mu)**2, dim=2) / self.beta) + 1e-8)
        q = q**(self.beta+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return q
    
    def target_distribution(self, q):
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def KL_div(self, p, q):
        loss = torch.mean(torch.sum(p*torch.log(p/(q+1e-6)), dim=1))
        return loss
  
    def forward(self, x, adj):
        z = self.gal(x, adj)
        q = self.get_q(z)
        return z, q

    def train_(self, X, adj, init="mclust",n_clusters=7,res=0.1,tol=1e-3,lr=0.1, 
                max_epochs=500, seed = 818, update_interval=3):

        optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=5e-4)
        
        X = torch.FloatTensor(X).to(self.device)
        adj = sparse_mx_to_torch_sparse_tensor(adj).to(self.device) 


        with torch.no_grad():
            features = self.gal(X, adj)
     
        #----------------------------------------------------------------           
        if init=="kmeans":
            print("\nInitializing cluster centers with kmeans, n_clusters known")
            self.n_clusters=n_clusters
            kmeans = KMeans(self.n_clusters, n_init=20)
            y_pred = kmeans.fit_predict(features.detach().cpu().numpy())
            
        elif init=="mclust":
            print("\nInitializing cluster centers with mclust, n_clusters known")
            data = features.detach().cpu().numpy()
            self.n_clusters = n_clusters
            self.seed = seed
            y_pred = mclust(data, num_cluster = self.n_clusters, random_seed = self.seed)
            y_pred = y_pred.astype(int)

        elif init=="louvain":
            print("\nInitializing cluster centers with louvain, resolution = ", res)
            adata = sc.AnnData(features.detach().cpu().numpy())
            sc.pp.neighbors(adata, n_neighbors=10)
            sc.tl.louvain(adata, resolution=res)
            y_pred = adata.obs['louvain'].astype(int).to_numpy()
            self.n_clusters = len(np.unique(y_pred))
            
        elif init=="leiden":
            print("\nInitializing cluster centers with leiden, resolution = ", res)
            adata=sc.AnnData(features.detach().cpu().numpy())
            sc.pp.neighbors(adata, n_neighbors=10)
            sc.tl.leiden(adata, resolution=res)
            y_pred = adata.obs['leiden'].astype(int).to_numpy()
            self.n_clusters = len(np.unique(y_pred))
        #----------------------------------------------------------------
        y_pred_last = y_pred
        self.mu = Parameter(torch.Tensor(self.n_clusters, self.embedding_size))

        features = pd.DataFrame(features.detach().cpu().numpy()).reset_index(drop = True)
        Group = pd.Series(y_pred, index=np.arange(0,features.shape[0]), name="Group")
        Mergefeature = pd.concat([features,Group], axis=1)
        cluster_centers = np.asarray(Mergefeature.groupby("Group").mean())       
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.mu.data = self.mu.data.to(self.device)
        
        #----------------------------------------------------------------
        
        with tqdm(total=max_epochs) as t:
            for epoch in range(max_epochs):            
                t.set_description('Epoch')
                self.train()
                
                if epoch%update_interval == 0:
                    _, Q = self.forward(X, adj)
                    q = Q.detach().data.cpu().numpy().argmax(1)              
                    t.update(update_interval)
                    
                z,q = self(X, adj)
                p = self.target_distribution(Q.detach())
                
                loss = self.KL_div(p, q)
   
                optimizer.zero_grad()              
                loss.backward()
                optimizer.step()
    
                t.set_postfix(loss = loss.data.cpu().numpy())
                
                #Check stop criterion
                y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / X.shape[0]
                y_pred_last = y_pred
                if epoch>0 and (epoch-1)%update_interval == 0 and delta_label < tol:
                    print('delta_label ', delta_label, '< tol ', tol)
                    print("Reach tolerance threshold. Stopping training.")
                    print("Total epoch:", epoch)
                    break

    def predict(self, X, adj):
        X = torch.FloatTensor(X).to(self.device)
        adj = sparse_mx_to_torch_sparse_tensor(adj).to(self.device)
        z, q = self(X, adj)

        return z, q
    


r"""
The downanlysis of cellphonedb
"""

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.patches as mpatches
import scanpy as sc
import matplotlib
import anndata

kpy_install=False

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

def check_kpy():
    r"""Check if ktplotspy package is installed for CellPhoneDB analysis.
    
    Raises:
        ImportError: If ktplotspy is not installed
    """
    global kpy_install
    try:
        import ktplotspy as kpy
        kpy_install=True
        print('ktplotspy have been install version:',kpy.__version__)
    except ImportError:
        raise ImportError(
            'Please install the ktplotspy: `pip install ktplotspy`.'
        )

def cpdb_network_cal(adata:anndata.AnnData,pvals:list,celltype_key:str)->dict:
    r"""
    Calculate a CPDB (Cell Phone Database) network using gene expression data and return a dictionary of results.

    Arguments:
        adata: An AnnData object containing gene expression data.
        pvals: A list or array of p-values for each gene in the expression data.
        celltype_key: The name of the cell type key in adata.obs.

    Returns:
        cpdb_dict: A dictionary of results generated by `kpy.plot_cpdb_heatmap()` from the `ktplotspy` library.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    cpdb_dict=kpy.plot_cpdb_heatmap(
        adata = adata,
        pvals = pvals,
        celltype_key = celltype_key,
        figsize = (1,1),
        title = "",
        symmetrical = False,
        return_tables=True,
    )
    return cpdb_dict


def cpdb_plot_network(adata:anndata.AnnData,interaction_edges:pd.DataFrame,
                      celltype_key:str,nodecolor_dict=None,
                      edgeswidth_scale:int=10,nodesize_scale:int=1,
                      pos_scale:int=1,pos_size:int=10,figsize:tuple=(5,5),title:str='',
                      legend_ncol:int=3,legend_bbox:tuple=(1,0.2),legend_fontsize:int=10,
                     return_graph:bool=False)->nx.Graph:
    r"""
    Plot a network of interactions between cell types using gene expression data.

    Arguments:
        adata: An AnnData object containing gene expression data.
        interaction_edges: A DataFrame containing the edges of the network, including the source, target, and count of interactions.
        celltype_key: The name of the cell type key in adata.obs.
        nodecolor_dict: A dictionary mapping cell type names to node colors. If not provided, uses the colors specified in adata.uns.
        edgeswidth_scale: The scaling factor for edge width. Default is 10.
        nodesize_scale: The scaling factor for node size. Default is 1.
        pos_scale: The scaling factor for node positions. Default is 1.
        pos_size: The size of the node positions. Default is 10.
        figsize: The size of the plot. Default is (5,5).
        title: The title of the plot. Default is ''.
        legend_ncol: The number of columns in the legend. Default is 3.
        legend_bbox: The location of the legend. Default is (1,0.2).
        legend_fontsize: The font size of the legend. Default is 10.
        return_graph: If True, returns the network graph. Default is False.

    Returns:
        ax or G: The plot if `return_graph` is False, otherwise the network graph.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    #set Digraph of cellphonedb
    G=nx.DiGraph()
    for i in interaction_edges.index:
        G.add_edge(interaction_edges.loc[i,'SOURCE'],
                   interaction_edges.loc[i,'TARGET'],
                   weight=interaction_edges.loc[i,'COUNT'],)
    
    #set celltypekey's color
    if nodecolor_dict!=None:
        type_color_all=nodecolor_dict
    else:
        if '{}_colors'.format(celltype_key) in adata.uns:
            type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,adata.uns['{}_colors'.format(celltype_key)]))
        else:
            if len(adata.obs[celltype_key].cat.categories)>28:
                type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.default_102))
            else:
                type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.zeileis_28))
    
    #set G_nodes_dict
    nodes=[]
    G_degree=dict(G.degree(G.nodes()))


    G_nodes_dict={}
    links = []
    for i in G.edges:
        if i[0] not in G_nodes_dict.keys():
            G_nodes_dict[i[0]]=0
        if i[1] not in G_nodes_dict.keys():
            G_nodes_dict[i[1]]=0
        links.append({"source": i[0], "target": i[1]})
        weight=G.get_edge_data(i[0],i[1])['weight']
        G_nodes_dict[i[0]]+=weight
        G_nodes_dict[i[1]]+=weight
        
    #plot
    fig, ax = plt.subplots(figsize=figsize) 
    pos = nx.spring_layout(G, scale=pos_scale, k=(pos_size)/np.sqrt(G.order()))
    p=dict(G.nodes)

    nodesize=np.array([G_nodes_dict[u] for u in G.nodes()])/nodesize_scale
    nodecolos=[type_color_all[u] for u in G.nodes()]
    nx.draw_networkx_nodes(G, pos, nodelist=p,node_size=nodesize,node_color=nodecolos)

    edgewidth = np.array([G.get_edge_data(u, v)['weight'] for u, v in G.edges()])/edgeswidth_scale
    nx.draw_networkx_edges(G, pos,width=edgewidth)


    #label_options = {"ec": "white", "fc": "white", "alpha": 0.6}
    #nx.draw_networkx_labels(G, pos, font_size=10,) #bbox=label_options)
    plt.grid(False)
    plt.axis("off")
    plt.xlim(-2,2)
    plt.ylim(-2,1.5)

    labels = adata.obs[celltype_key].cat.categories
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    color = [type_color_all[u] for u in labels]
    patches = [mpatches.Patch(color=type_color_all[u], label=u) for u in labels ] 

    #plt.xlim(-0.05, 1.05)
    #plt.ylim(-0.05, 1.05)
    plt.axis("off")
    plt.title(title)
    plt.legend(handles=patches,
               bbox_to_anchor=legend_bbox, 
               ncol=legend_ncol,
               fontsize=legend_fontsize)
    if return_graph==True:
        return G
    else:
        return ax
    #return {'Graph':G,'ax':ax}

def cpdb_plot_interaction(adata:anndata.AnnData,cell_type1:str,cell_type2:str,
                          means:pd.DataFrame,pvals:pd.DataFrame,
                          celltype_key:str,genes=None,
                         keep_significant_only:bool=True,figsize:tuple = (4,8),title:str="",
                         max_size:int=1,highlight_size:float = 0.75,standard_scale:bool = True,
                         cmap_name:str='viridis',
                         ytickslabel_fontsize:int=8,xtickslabel_fontsize:int=8,title_fontsize:int=10)->matplotlib.axes._axes.Axes:
    r"""Plot CellPhoneDB cell-cell interactions between two cell types.

    Arguments:
        adata: AnnData object containing single-cell data
        cell_type1 (str): Name of first cell type
        cell_type2 (str): Name of second cell type
        means: DataFrame containing interaction means from CellPhoneDB
        pvals: DataFrame containing p-values from CellPhoneDB
        celltype_key (str): Key in adata.obs containing cell type information
        genes (list): List of genes to include in plot (default: None for all genes)
        keep_significant_only (bool): Whether to show only significant interactions (default: True)
        figsize (tuple): Figure size (default: (4,8))
        title (str): Plot title (default: '')
        max_size (int): Maximum dot size (default: 1)
        highlight_size (float): Size for highlighted interactions (default: 0.75)
        standard_scale (bool): Whether to standard scale the data (default: True)
        cmap_name (str): Name of the colormap to use (default: 'viridis')
        ytickslabel_fontsize (int): Font size of the y-axis tick labels (default: 8)
        xtickslabel_fontsize (int): Font size of the x-axis tick labels (default: 8)
        title_fontsize (int): Font size of the title (default: 10)
    
    Returns:
        matplotlib.axes.Axes: Axes object containing the interaction plot

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    fig=kpy.plot_cpdb(
        adata = adata,
        cell_type1 = cell_type1,
        cell_type2 = cell_type2, 
        means = means,
        pvals = pvals,
        celltype_key = celltype_key,
        genes = genes,
        keep_significant_only=keep_significant_only,
        figsize = figsize,
        title = "",
        max_size = max_size,
        highlight_size = highlight_size,
        standard_scale = standard_scale,
        cmap_name=cmap_name
    ).draw()
    
    #ytickslabels
    labels=fig.get_axes()[0].yaxis.get_ticklabels()
    plt.setp(labels, fontsize=ytickslabel_fontsize)

    #xtickslabels
    labels=fig.get_axes()[0].xaxis.get_ticklabels()
    plt.setp(labels, fontsize=xtickslabel_fontsize)

    fig.get_axes()[0].set_title(title,fontsize=title_fontsize)
    
    return fig.get_axes()[0]

def cpdb_submeans_exacted(means:pd.DataFrame,cell_names:str,cell_type:str='ligand')->pd.DataFrame:
    r"""Extract subset of CellPhoneDB means DataFrame for specific cell type.

    Arguments:
        means: DataFrame containing interaction means from CellPhoneDB
        cell_names (str): Name of the cell type to extract
        cell_type (str): Whether to extract as 'ligand' or 'receptor' (default: 'ligand')

    Returns:
        pd.DataFrame: Subset of means DataFrame for specified cell type and role

    """
    if cell_type=='ligand':
        means_columns=means.columns[:11].tolist()+means.columns[means.columns.str.contains('{}\|'.format(cell_names))].tolist()
    elif cell_type=='receptor':
        means_columns=means.columns[:11].tolist()+means.columns[means.columns.str.contains('\|{}'.format(cell_names))].tolist()
    else:
        raise ValueError('cell_type must be ligand or receptor')
    return means.loc[:,means_columns]

def cpdb_interaction_filtered(adata:anndata.AnnData,cell_type1:str,cell_type2:str,
                              means:pd.DataFrame,pvals:pd.DataFrame,celltype_key:str,genes=None,
                         keep_significant_only:bool=True,figsize:tuple = (0,0),
                         max_size:int=1,highlight_size:float = 0.75,standard_scale:bool = True,cmap_name:str='viridis',)->list:
    r"""
    Returns a list of unique interaction groups from a CellPhoneDB analysis filtered by significance.

    Arguments:
        adata: Annotated data matrix containing normalized gene expression data.
        cell_type1: The cell type name for the first interacting partner.
        cell_type2: The cell type name for the second interacting partner.
        means: Dataframe containing the means of gene expression for each interaction.
        pvals: Dataframe containing the p-values for each interaction.
        celltype_key: The column name in the `adata.obs` dataframe containing cell type annotations.
        genes: List of gene names to filter on. If None, no filtering is applied.
        keep_significant_only: Whether to filter on significant interactions only (i.e. interactions with FDR < 0.05).
        figsize: Figure size. Width and height of the figure in inches. 
        max_size: Maximum size of the markers in the plot.
        highlight_size: Size of the markers for the highlighted interaction.
        standard_scale: Whether to standard scale the data.
        cmap_name: Name of the colormap to use.

    Returns:
        genes: List of unique interaction groups from a CellPhoneDB analysis filtered by significance.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    res=kpy.plot_cpdb(
        adata = adata,
        cell_type1 = cell_type1,
        cell_type2 = cell_type2, 
        means = means,
        pvals = pvals,
        celltype_key = celltype_key,
        genes = genes,
        keep_significant_only=keep_significant_only,
        figsize = figsize,
        title = "",
        max_size = max_size,
        highlight_size = highlight_size,
        standard_scale = standard_scale,
        cmap_name=cmap_name,
        return_table=True
    )

    return list(set(res['interaction_group']))

def cpdb_exact_target(means,target_cells):
    import re
    
    t_dict=[]
    for t in target_cells:
        escaped_str = re.escape('|'+t)
        target_names=means.columns[means.columns.str.contains(escaped_str)].tolist()
        t_dict+=target_names
    #print(t_dict)
    target_sub=means[means.columns[:10].tolist()+t_dict]
    return target_sub

def cpdb_exact_source(means,source_cells):
    import re
    
    t_dict=[]
    for t in source_cells:
        escaped_str = re.escape(t+'|')
        source_names=means.columns[means.columns.str.contains(escaped_str)].tolist()
        t_dict+=source_names
    #print(t_dict)
    source_sub=means[means.columns[:10].tolist()+t_dict]
    return source_sub


from tqdm import tqdm
def cpdb2cellchat(df):
    new_columns = ['source', 'target', 'ligand', 'receptor', 'prob', 'pval', 'interaction_name', 'interaction_name_2', 'pathway_name', 'annotation', 'evidence']
    new_df = pd.DataFrame(columns=new_columns)
    
    # 遍历每一行和细胞对列
    for index, row in tqdm(df.iterrows()):
        for col in df.columns[14:]:  # 假设从第14列开始是细胞对
            if pd.notna(row[col]):
                source, target = col.split('|')  # 通过 '|' 分割 source 和 target
                if pd.notna(row['gene_a']) and pd.notna(row['gene_b']):
                    interaction_name=row['interacting_pair'].split('_')
                    if len(interaction_name)>2:
                        interaction_name_2=row['interacting_pair'].split('_')[0]+' - ('+'+'.join(row['interacting_pair'].split('_')[1:])+')'
                    else:
                        interaction_name_2=row['interacting_pair'].split('_')[0]+' - '+'+'.join(row['interacting_pair'].split('_')[1:])
                    new_row = {
                        'source': source,
                        'target': target,
                        'ligand': row['gene_a'],
                        'receptor': row['gene_b'],
                        'prob': row['rank'],  # 假设 rank 是概率
                        'pval': 0,
                        'interaction_name': row['interacting_pair'],
                        'interaction_name_2': interaction_name_2,  # 假设相同
                        'pathway_name': row['classification'],
                        'annotation': row['annotation_strategy'],
                        'evidence': 'curated'  # 假设证据为 curated
                    }
                    new_df = new_df.append(new_row, ignore_index=True)
    
    # 显示新的 DataFrame
    return new_df


def cellphonedb_v5(adata, 
                           celltype_key='celltype',
                           min_cell_fraction=0.005,
                           min_genes=200,
                           min_cells=3,
                           cpdb_file_path=None,
                           iterations=1000,
                           threshold=0.1,
                           pvalue=0.05,
                           threads=10,
                           output_dir=None,
                           temp_dir=None,
                           cleanup_temp=True,
                           debug=False,
                           separator='|',
                           **kwargs):
    """
    Run CellPhoneDB statistical analysis with proper file handling
    
    Parameters:
    -----------
    adata : AnnData
        Annotated data matrix
    celltype_key : str
        Column name in adata.obs containing cell type annotations
    min_cell_fraction : float
        Minimum fraction of total cells required for a cell type to be included
    min_genes : int
        Minimum number of genes required per cell
    min_cells : int
        Minimum number of cells required per gene
    cpdb_file_path : str or None
        Path to CellPhoneDB database zip file. If None, will try to find automatically
    iterations : int
        Number of shufflings performed in the analysis
    threshold : float
        Min % of cells expressing a gene for this to be employed in the analysis
    pvalue : float
        P-value threshold to employ for significance
    threads : int
        Number of threads to use in the analysis
    output_dir : str or None
        Directory to save results. If None, creates temporary directory
    temp_dir : str or None
        Directory for temporary files. If None, uses system temp
    cleanup_temp : bool
        Whether to clean up temporary files after analysis
    debug : bool
        Saves all intermediate tables employed during the analysis
    separator : str
        String to employ to separate cells in the results dataframes
    **kwargs : dict
        Additional parameters for cpdb_statistical_analysis_method.call
        
    Returns:
    --------
    dict : CellPhoneDB results
    adata_cpdb : AnnData
        Formatted AnnData object for visualization with CellChatViz
    """
    import os
    import tempfile
    import shutil
    from pathlib import Path
    import pandas as pd
    import scanpy as sc
    
    # Validate inputs
    if celltype_key not in adata.obs.columns:
        raise ValueError(f"celltype_key '{celltype_key}' not found in adata.obs")
    
    print("🔬 Starting CellPhoneDB analysis...")
    print(f"   - Original data: {adata.shape[0]} cells, {adata.shape[1]} genes")
    
    # Step 1: Filter cell types by minimum cell fraction
    ct_counts = adata.obs[celltype_key].value_counts()
    min_cells_required = int(adata.shape[0] * min_cell_fraction)
    valid_celltypes = ct_counts[ct_counts > min_cells_required].index
    
    print(f"   - Cell types passing {min_cell_fraction*100}% threshold: {len(valid_celltypes)}")
    print(f"   - Minimum cells required: {min_cells_required}")
    
    if len(valid_celltypes) == 0:
        raise ValueError("No cell types pass the minimum cell fraction threshold")
    
    # Step 2: Subset and preprocess data
    adata_filtered = adata[adata.obs[celltype_key].isin(valid_celltypes)].copy()
    
    # Use raw data if available, otherwise use X
    if adata_filtered.raw is not None:
        adata_pp = adata_filtered.raw.to_adata()
        adata_pp.obs = adata_filtered.obs.copy()
    else:
        adata_pp = adata_filtered.copy()
    
    print(f"   - After filtering: {adata_pp.shape[0]} cells, {adata_pp.shape[1]} genes")
    
    # Apply standard preprocessing
    sc.pp.filter_cells(adata_pp, min_genes=min_genes)
    sc.pp.filter_genes(adata_pp, min_cells=min_cells)
    
    print(f"   - After preprocessing: {adata_pp.shape[0]} cells, {adata_pp.shape[1]} genes")
    
    # Step 3: Setup directories
    if temp_dir is None:
        temp_dir = tempfile.mkdtemp(prefix='cpdb_temp_')
        temp_created = True
    else:
        temp_created = False
        os.makedirs(temp_dir, exist_ok=True)
    
    if output_dir is None:
        output_dir = tempfile.mkdtemp(prefix='cpdb_results_')
        output_created = True
    else:
        output_created = False
        os.makedirs(output_dir, exist_ok=True)
    
    print(f"   - Temporary directory: {temp_dir}")
    print(f"   - Output directory: {output_dir}")
    
    try:
        # Step 4: Create temporary files for CellPhoneDB
        # Create counts file (AnnData format)
        counts_file = os.path.join(temp_dir, 'counts_matrix.h5ad')
        adata_counts = sc.AnnData(
            X=adata_pp.X,
            obs=pd.DataFrame(index=adata_pp.obs.index),
            var=pd.DataFrame(index=adata_pp.var.index)
        )
        adata_counts.write_h5ad(counts_file, compression='gzip')
        
        # Create metadata file
        meta_file = os.path.join(temp_dir, 'metadata.tsv')
        df_meta = pd.DataFrame({
            'Cell': list(adata_pp.obs.index),
            'cell_type': list(adata_pp.obs[celltype_key])
        })
        df_meta.set_index('Cell', inplace=True)
        df_meta.to_csv(meta_file, sep='\t')
        
        print("   - Created temporary input files")
        
        # Step 5: Find CellPhoneDB database if not provided
        if cpdb_file_path is None:
            # Try common locations
            possible_paths = [
                '/oak/stanford/groups/xiaojie/steorra/software/cellphonedb.zip',
                './cellphonedb.zip',
                '~/cellphonedb.zip'
            ]
            
            for path in possible_paths:
                expanded_path = os.path.expanduser(path)
                if os.path.exists(expanded_path):
                    cpdb_file_path = expanded_path
                    break
            
            if cpdb_file_path is None:
                raise FileNotFoundError(
                    "CellPhoneDB database not found. Please provide cpdb_file_path or "
                    "place cellphonedb.zip in current directory"
                )
        
        print(f"   - Using CellPhoneDB database: {cpdb_file_path}")
        
        # Step 6: Run CellPhoneDB analysis
        try:
            from cellphonedb.src.core.methods import cpdb_statistical_analysis_method
        except ImportError:
            raise ImportError(
                "CellPhoneDB not installed. Please install with: "
                "pip install cellphonedb"
            )
        
        print("   - Running CellPhoneDB statistical analysis...")
        
        # Prepare parameters
        analysis_params = {
            'cpdb_file_path': cpdb_file_path,
            'meta_file_path': meta_file,
            'counts_file_path': counts_file,
            'counts_data': 'hgnc_symbol',
            'active_tfs_file_path': None,
            'microenvs_file_path': None,
            'score_interactions': True,
            'iterations': iterations,
            'threshold': threshold,
            'threads': threads,
            'debug_seed': 42,
            'result_precision': 3,
            'pvalue': pvalue,
            'subsampling': False,
            'subsampling_log': False,
            'subsampling_num_pc': 100,
            'subsampling_num_cells': 1000,
            'separator': separator,
            'debug': debug,
            'output_path': output_dir,
            'output_suffix': None
        }
        
        # Update with any additional kwargs
        analysis_params.update(kwargs)
        
        # Run analysis
        cpdb_results = cpdb_statistical_analysis_method.call(**analysis_params)
        
        print("   - CellPhoneDB analysis completed successfully!")
        
        # Step 7: Format results for visualization
        print("   - Formatting results for visualization...")
        
        adata_cpdb = format_cpdb_results_for_viz(cpdb_results, separator=separator)
        
        print(f"   - Created visualization AnnData: {adata_cpdb.shape}")
        print(f"   - Cell interactions: {adata_cpdb.n_obs}")
        print(f"   - L-R pairs: {adata_cpdb.n_vars}")
        
        return cpdb_results, adata_cpdb
        
    finally:
        # Step 8: Cleanup temporary files if requested
        if cleanup_temp:
            if temp_created and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)
                print(f"   - Cleaned up temporary directory: {temp_dir}")
            
            if output_created and os.path.exists(output_dir):
                # Only clean output if we created it and user didn't specify it
                pass  # Keep results by default
        
        print("✅ CellPhoneDB analysis pipeline completed!")


def format_cpdb_results_for_viz(cpdb_results, separator='|'):
    """
    Format CellPhoneDB results into AnnData object for CellChatViz
    
    Parameters:
    -----------
    cpdb_results : dict
        Results from CellPhoneDB analysis
    separator : str
        Separator used in cell type pairs
        
    Returns:
    --------
    adata_cpdb : AnnData
        Formatted AnnData object for visualization
    """
    import pandas as pd
    import scanpy as sc
    
    # Extract results
    means_df = cpdb_results['means']
    pvalues_df = cpdb_results['pvalues']
    
    # Identify cell type pair columns (usually start after column 12 or 13)
    info_cols = []
    pair_cols = []
    
    for col in means_df.columns:
        if separator in str(col):
            pair_cols.append(col)
        else:
            info_cols.append(col)
    
    print(f"   - Found {len(info_cols)} info columns and {len(pair_cols)} cell type pairs")
    
    if len(pair_cols) == 0:
        raise ValueError(f"No cell type pair columns found with separator '{separator}'")
    
    # Create AnnData object
    # X matrix: cell pairs (obs) x L-R interactions (vars)
    X_data = means_df[pair_cols].T  # Transpose so pairs are observations
    
    adata_cpdb = sc.AnnData(X=X_data)
    
    # Add layers
    adata_cpdb.layers['means'] = means_df[pair_cols].T
    adata_cpdb.layers['pvalues'] = pvalues_df[pair_cols].T
    
    # Add variable (L-R pair) information
    adata_cpdb.var = means_df[info_cols].copy()
    adata_cpdb.var['interaction_name'] = adata_cpdb.var['interacting_pair']
    
    # Add observation (cell pair) information
    adata_cpdb.obs['sender'] = [pair.split(separator)[0] for pair in adata_cpdb.obs.index]
    adata_cpdb.obs['receiver'] = [pair.split(separator)[1] for pair in adata_cpdb.obs.index]
    
    # Add interaction classification if available
    if 'classification' in adata_cpdb.var.columns:
        print(f"   - Found {adata_cpdb.var['classification'].nunique()} pathway classifications")
    
    return adata_cpdb


def create_cellchatviz_from_cpdb(cpdb_results, separator='|', palette=None):
    """
    Create CellChatViz object directly from CellPhoneDB results
    
    Parameters:
    -----------
    cpdb_results : dict
        Results from CellPhoneDB analysis
    separator : str
        Separator used in cell type pairs
    palette : dict or list, optional
        Color palette for cell types
        
    Returns:
    --------
    viz : CellChatViz
        Initialized CellChatViz object ready for visualization
    """
    # Format results
    adata_cpdb = format_cpdb_results_for_viz(cpdb_results, separator=separator)
    
    # Create and return CellChatViz object
    from ..pl._cpdbviz import CellChatViz
    viz = CellChatViz(adata_cpdb, palette=palette)
    
    print(f"✅ Created CellChatViz with {viz.n_cell_types} cell types")
    
    return viz


def download_cellphonedb_database(download_path=None, force_download=False):
    """
    Download CellPhoneDB database with fallback URLs
    
    Parameters:
    -----------
    download_path : str or None
        Path to save the database. If None, saves to current directory as 'cellphonedb.zip'
    force_download : bool
        Whether to force download even if file exists
        
    Returns:
    --------
    str : Path to downloaded database file
    """
    import os
    import urllib.request
    import urllib.error
    from pathlib import Path
    
    if download_path is None:
        download_path = './cellphonedb.zip'
    
    download_path = Path(download_path)
    
    # Check if file already exists
    if download_path.exists() and not force_download:
        print(f"✅ CellPhoneDB database already exists: {download_path}")
        return str(download_path)
    
    # URLs to try in order
    download_urls = [
        "https://github.com/ventolab/cellphonedb-data/raw/refs/heads/master/cellphonedb.zip",
        "https://starlit.oss-cn-beijing.aliyuncs.com/single/cellphonedb.zip"
    ]
    
    print("📥 Downloading CellPhoneDB database...")
    
    for i, url in enumerate(download_urls, 1):
        try:
            print(f"   - Trying URL {i}/{len(download_urls)}: {url}")
            
            # Create directory if it doesn't exist
            download_path.parent.mkdir(parents=True, exist_ok=True)
            
            # Download with progress
            def show_progress(block_num, block_size, total_size):
                if total_size > 0:
                    percent = min(100, block_num * block_size * 100 / total_size)
                    print(f"\r     Progress: {percent:.1f}%", end='', flush=True)
            
            urllib.request.urlretrieve(url, download_path, reporthook=show_progress)
            print(f"\n✅ Successfully downloaded CellPhoneDB database to: {download_path}")
            
            # Verify file is not empty
            if download_path.stat().st_size > 0:
                return str(download_path)
            else:
                print(f"❌ Downloaded file is empty, trying next URL...")
                download_path.unlink(missing_ok=True)
                
        except urllib.error.URLError as e:
            print(f"\n❌ Failed to download from {url}: {e}")
            download_path.unlink(missing_ok=True)
            continue
        except Exception as e:
            print(f"\n❌ Unexpected error downloading from {url}: {e}")
            download_path.unlink(missing_ok=True)
            continue
    
    raise RuntimeError(
        "Failed to download CellPhoneDB database from all available URLs. "
        "Please check your internet connection or manually download the database."
    )


def validate_cpdb_database(cpdb_file_path):
    """
    Validate CellPhoneDB database file
    
    Parameters:
    -----------
    cpdb_file_path : str
        Path to CellPhoneDB database file
        
    Returns:
    --------
    str : Validated database path
    """
    import os
    from pathlib import Path
    
    if cpdb_file_path is None:
        raise ValueError("cpdb_file_path is required. Use download_cellphonedb_database() to get the database.")
    
    cpdb_path = Path(cpdb_file_path)
    
    # If path doesn't exist, try to download
    if not cpdb_path.exists():
        print(f"❌ Database not found at: {cpdb_file_path}")
        print("🔄 Attempting to download database...")
        
        # If the provided path looks like a filename, use it for download
        if cpdb_path.name.endswith('.zip'):
            return download_cellphonedb_database(cpdb_file_path)
        else:
            # Download to default location and return that path
            downloaded_path = download_cellphonedb_database()
            print(f"💡 Database downloaded to: {downloaded_path}")
            print(f"   You can use this path in future calls: cpdb_file_path='{downloaded_path}'")
            return downloaded_path
    
    # Validate file size
    file_size = cpdb_path.stat().st_size
    if file_size == 0:
        print(f"❌ Database file is empty: {cpdb_file_path}")
        print("🔄 Re-downloading database...")
        return download_cellphonedb_database(cpdb_file_path, force_download=True)
    
    print(f"✅ Valid CellPhoneDB database found: {cpdb_file_path} ({file_size/1024/1024:.1f} MB)")
    return str(cpdb_path)


def run_cellphonedb_v5(adata, 
                           cpdb_file_path,  # Now mandatory
                           celltype_key='celltype',
                           min_cell_fraction=0.005,
                           min_genes=200,
                           min_cells=3,
                           iterations=1000,
                           threshold=0.1,
                           pvalue=0.05,
                           threads=10,
                           output_dir=None,
                           temp_dir=None,
                           cleanup_temp=True,
                           debug=False,
                           separator='|',
                           **kwargs):
    """
    Run CellPhoneDB statistical analysis with automatic database download
    
    Parameters:
    -----------
    adata : AnnData
        Annotated data matrix
    cpdb_file_path : str
        Path to CellPhoneDB database zip file (REQUIRED)
        If file doesn't exist, will attempt automatic download
    celltype_key : str
        Column name in adata.obs containing cell type annotations
    min_cell_fraction : float
        Minimum fraction of total cells required for a cell type to be included
    min_genes : int
        Minimum number of genes required per cell
    min_cells : int
        Minimum number of cells required per gene
    iterations : int
        Number of shufflings performed in the analysis
    threshold : float
        Min % of cells expressing a gene for this to be employed in the analysis
    pvalue : float
        P-value threshold to employ for significance
    threads : int
        Number of threads to use in the analysis
    output_dir : str or None
        Directory to save results. If None, creates temporary directory
    temp_dir : str or None
        Directory for temporary files. If None, uses system temp
    cleanup_temp : bool
        Whether to clean up temporary files after analysis
    debug : bool
        Saves all intermediate tables employed during the analysis
    separator : str
        String to employ to separate cells in the results dataframes
    **kwargs : dict
        Additional parameters for cpdb_statistical_analysis_method.call
        
    Returns:
    --------
    dict : CellPhoneDB results
    adata_cpdb : AnnData
        Formatted AnnData object for visualization with CellChatViz
        
    Examples:
    --------
    # Basic usage - will download database automatically if needed
    cpdb_results, adata_cpdb = run_cellphonedb_analysis(
        adata, 
        cpdb_file_path='./cellphonedb.zip',
        celltype_key='celltype_minor'
    )
    
    # Advanced usage
    cpdb_results, adata_cpdb = run_cellphonedb_analysis(
        adata,
        cpdb_file_path='/path/to/cellphonedb.zip',
        celltype_key='celltype_minor',
        min_cell_fraction=0.01,
        iterations=2000,
        threads=20
    )
    """
    import os
    import tempfile
    import shutil
    from pathlib import Path
    import pandas as pd
    import scanpy as sc
    
    # Step 1: Validate and download database if necessary
    print("🔬 Starting CellPhoneDB analysis...")
    cpdb_file_path = validate_cpdb_database(cpdb_file_path)
    
    # Validate inputs
    if celltype_key not in adata.obs.columns:
        raise ValueError(f"celltype_key '{celltype_key}' not found in adata.obs")
    
    print(f"   - Original data: {adata.shape[0]} cells, {adata.shape[1]} genes")
    
    # Step 2: Filter cell types by minimum cell fraction
    ct_counts = adata.obs[celltype_key].value_counts()
    min_cells_required = int(adata.shape[0] * min_cell_fraction)
    valid_celltypes = ct_counts[ct_counts > min_cells_required].index
    
    print(f"   - Cell types passing {min_cell_fraction*100}% threshold: {len(valid_celltypes)}")
    print(f"   - Minimum cells required: {min_cells_required}")
    
    if len(valid_celltypes) == 0:
        raise ValueError("No cell types pass the minimum cell fraction threshold")
    
    # Step 3: Subset and preprocess data
    adata_filtered = adata[adata.obs[celltype_key].isin(valid_celltypes)].copy()
    
    # Use raw data if available, otherwise use X
    if adata_filtered.raw is not None:
        adata_pp = adata_filtered.raw.to_adata()
        adata_pp.obs = adata_filtered.obs.copy()
    else:
        adata_pp = adata_filtered.copy()
    
    print(f"   - After filtering: {adata_pp.shape[0]} cells, {adata_pp.shape[1]} genes")
    
    # Apply standard preprocessing
    sc.pp.filter_cells(adata_pp, min_genes=min_genes)
    sc.pp.filter_genes(adata_pp, min_cells=min_cells)
    
    print(f"   - After preprocessing: {adata_pp.shape[0]} cells, {adata_pp.shape[1]} genes")
    
    # Step 4: Setup directories
    if temp_dir is None:
        temp_dir = tempfile.mkdtemp(prefix='cpdb_temp_')
        temp_created = True
    else:
        temp_created = False
        os.makedirs(temp_dir, exist_ok=True)
    
    if output_dir is None:
        output_dir = tempfile.mkdtemp(prefix='cpdb_results_')
        output_created = True
    else:
        output_created = False
        os.makedirs(output_dir, exist_ok=True)
    
    print(f"   - Temporary directory: {temp_dir}")
    print(f"   - Output directory: {output_dir}")
    
    try:
        # Step 5: Create temporary files for CellPhoneDB
        # Create counts file (AnnData format)
        counts_file = os.path.join(temp_dir, 'counts_matrix.h5ad')
        adata_counts = sc.AnnData(
            X=adata_pp.X,
            obs=pd.DataFrame(index=adata_pp.obs.index),
            var=pd.DataFrame(index=adata_pp.var.index)
        )
        adata_counts.write_h5ad(counts_file, compression='gzip')
        
        # Create metadata file
        meta_file = os.path.join(temp_dir, 'metadata.tsv')
        df_meta = pd.DataFrame({
            'Cell': list(adata_pp.obs.index),
            'cell_type': list(adata_pp.obs[celltype_key])
        })
        df_meta.set_index('Cell', inplace=True)
        df_meta.to_csv(meta_file, sep='\t')
        
        print("   - Created temporary input files")
        
        # Step 6: Run CellPhoneDB analysis
        try:
            from cellphonedb.src.core.methods import cpdb_statistical_analysis_method
        except ImportError:
            raise ImportError(
                "CellPhoneDB not installed. Please install with: "
                "pip install cellphonedb"
            )
        
        print("   - Running CellPhoneDB statistical analysis...")
        
        # Prepare parameters
        analysis_params = {
            'cpdb_file_path': cpdb_file_path,
            'meta_file_path': meta_file,
            'counts_file_path': counts_file,
            'counts_data': 'hgnc_symbol',
            'active_tfs_file_path': None,
            'microenvs_file_path': None,
            'score_interactions': True,
            'iterations': iterations,
            'threshold': threshold,
            'threads': threads,
            'debug_seed': 42,
            'result_precision': 3,
            'pvalue': pvalue,
            'subsampling': False,
            'subsampling_log': False,
            'subsampling_num_pc': 100,
            'subsampling_num_cells': 1000,
            'separator': separator,
            'debug': debug,
            'output_path': output_dir,
            'output_suffix': None
        }
        
        # Update with any additional kwargs
        analysis_params.update(kwargs)
        
        # Run analysis
        cpdb_results = cpdb_statistical_analysis_method.call(**analysis_params)
        
        print("   - CellPhoneDB analysis completed successfully!")
        
        # Step 7: Format results for visualization
        print("   - Formatting results for visualization...")
        
        adata_cpdb = format_cpdb_results_for_viz(cpdb_results, separator=separator)
        
        print(f"   - Created visualization AnnData: {adata_cpdb.shape}")
        print(f"   - Cell interactions: {adata_cpdb.n_obs}")
        print(f"   - L-R pairs: {adata_cpdb.n_vars}")
        
        return cpdb_results, adata_cpdb
        
    finally:
        # Step 8: Cleanup temporary files if requested
        if cleanup_temp:
            if temp_created and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)
                print(f"   - Cleaned up temporary directory: {temp_dir}")
            
            if output_created and os.path.exists(output_dir):
                # Only clean output if we created it and user didn't specify it
                pass  # Keep results by default
        
        print("✅ CellPhoneDB analysis pipeline completed!")


import re
import sys
import os
from collections import defaultdict
from subprocess import Popen,PIPE

def train_NN_parallel(path_to_coords, path_to_glmpca, hidden_spatial, hidden_expression, 
                      output_dir, conda_environment, path_to_conda_folder,
                      epochs=10000, checkpoint=500, optimizer='adam', num_seeds=30, partition=None,
                     time="0-01:00:00"):

    hidden_spatial=' '.join(map(str, hidden_spatial))
    hidden_expression=' '.join(map(str, hidden_expression))
    
    tasks = 1
    mem_per_cpu = 5000
    processes = 1
    for seed in range(num_seeds):
        os.makedirs(f"{output_dir}/seed{seed}", exist_ok=True)
        cmd = f"gaston -i {path_to_coords} -o {path_to_glmpca} "
        cmd += f"--epochs {epochs} -d {output_dir} "
        cmd += f"--hidden_spatial {hidden_spatial} --hidden_expression {hidden_expression} "
        cmd += f"--optimizer {optimizer} --seed {seed} -c {checkpoint}"
        create_job_script(seed, f"{output_dir}/seed{seed}", tasks, processes, time, mem_per_cpu, cmd, conda_environment,path_to_conda_folder, partition=partition)

def create_job_script(name, outDir, tasks, cpuPerTask, time, mem_per_cpu, command, environment, path_to_conda_folder, partition=None):
    outFile = open('job_%s.sh' % name , 'w')
    o = outDir + "/out." + str(name)
    e = outDir + "/err." + str(name)
    print("#!/bin/bash", file=outFile)
    print("#SBATCH -J "+ str(name), file=outFile)
    print("#SBATCH -o " + o, file=outFile)
    print("#SBATCH -e " + e, file=outFile)
    print("#SBATCH --nodes=1", file=outFile)
    print("#SBATCH --ntasks=" + str(tasks), file=outFile)
    print("#SBATCH --cpus-per-task=" + str(cpuPerTask), file=outFile)
    print("#SBATCH -t " + str(time), file=outFile)
    print("#SBATCH --mem-per-cpu=" + str(mem_per_cpu), file=outFile)
    print(f"source {path_to_conda_folder} base", file=outFile)
    print(f"conda activate {environment}", file=outFile)
    print(command, file=outFile)
    outFile.close()
    jobId = sbatch_submit(outFile.name, partition)
    print(f'jobId: {jobId}')
    os.system("mv job_" + str(name) + ".sh " + outDir)
    return(jobId)

#Submit filename to slurm with sbatch, returns job id number
def sbatch_submit(filename, partition):
    print(f'partition: {partition}')
    if partition is not None:
        proc=Popen(f'sbatch -A {partition} {filename}',shell=True,stdout=PIPE,stderr=PIPE)
    else:
        proc = Popen('sbatch %s'%filename,shell=True,stdout=PIPE,stderr=PIPE)
    stdout,stderr = proc.communicate()
    stdout = stdout.decode("utf-8","ignore")
    stdout = stdout.strip()
    stdout = stdout.strip('Submitted batch job ')
    return(stdout)
